import typing
import dataclasses
import utils
import sys
import myTypeguard
import errors
import typecheck
import location
import stacktrace
from utils import _call_with_frames_removed

EQ_ATTRS_ATTR = '__eqAttrs__'

_typeCheckingEnabled = False

def init(enableTypeChecking=True):
    global _typeCheckingEnabled
    _typeCheckingEnabled = enableTypeChecking

def _collectDataClassAttributes(cls):
    result = dict()
    for c in cls.mro():
        if hasattr(c, '__kind') and c.__kind == 'record' and hasattr(c, '__annotations__'):
            result = c.__annotations__ | result
    return result

def _checkRecordAttr(cls: typing.Any,
                     ns: myTypeguard.Namespaces,
                     name: str,
                     ty: typing.Any,
                     tyLoc: typing.Optional[location.Loc],
                     v: typing.Any):
    if not typecheck.handleMatchesTyResult(myTypeguard.matchesTy(v, ty, ns), tyLoc):
        fi = stacktrace.callerOutsideWypp()
        if fi:
            loc = location.Loc.fromFrameInfo(fi)
        else:
            loc = None
        raise errors.WyppTypeError.recordAssignError(cls.__name__,
                                                        name,
                                                        ty,
                                                        tyLoc,
                                                        v,
                                                        loc)
    return v

def _patchDataClass(cls, mutable: bool, ns: myTypeguard.Namespaces):
    fieldNames = [f.name for f in dataclasses.fields(cls)]
    setattr(cls, EQ_ATTRS_ATTR, fieldNames)

    if hasattr(cls, '__annotations__'):
        # add annotions for type checked constructor.
        cls.__kind = 'record'
        cls.__init__.__annotations__ = _collectDataClassAttributes(cls)
        cls.__init__ = typecheck.wrapTypecheckRecordConstructor(cls, ns)

    if mutable:
        # prevent new fields being added
        fields = set(fieldNames)
        info = location.RecordConstructorInfo(cls)
        locs = {}
        for name in fields:
            if not name in cls.__annotations__:
                raise errors.WyppTypeError.noTypeAnnotationForRecordAttribute(name, cls.__name__)
            else:
                locs[name] = info.getParamSourceLocation(name)
        types = None # lazily initialized because forward references are not available at this point
        oldSetattr = cls.__setattr__
        def _setattr(obj, name, v):
            nonlocal types
            if types is None:
                types = typing.get_type_hints(cls, globalns=ns.globals, localns=ns.locals,
                                              include_extras=True)
            if name in types:
                ty = types[name]
                tyLoc = locs[name]
                v = _checkRecordAttr(cls, ns, name, ty, tyLoc, v)
                oldSetattr(obj, name, v)
            else:
                raise errors.WyppAttributeError(f'Unknown attribute {name} for record {cls.__name__}')
        setattr(cls, "__setattr__", lambda obj, k, v: _call_with_frames_removed(_setattr, obj, k, v))
    return cls

def record(cls=None, mutable=False, globals={}, locals={}):
    ns = myTypeguard.Namespaces(globals, locals)
    def wrap(cls: type):
        newCls = dataclasses.dataclass(cls, frozen=not mutable)
        if _typeCheckingEnabled:
            return utils._call_with_frames_removed(_patchDataClass, newCls, mutable, ns)
        else:
            return newCls
    # See if we're being called as @record or @record().
    if cls is None:
        # We're called with parens.
        return wrap
    else:
        # We're called as @dataclass without parens.
        return _call_with_frames_removed(wrap, cls)



import utils
import sys

VERBOSE = False # set via commandline
DEBUG = utils.getEnv("WYPP_DEBUG", bool, False)

def isDebug() -> bool:
    return DEBUG

def enableVerbose():
    global VERBOSE
    VERBOSE = True

def enableDebug():
    global VERBOSE, DEBUG
    VERBOSE = True
    DEBUG = True

def printStderr(s=''):
    sys.stderr.write(s + '\n')

def verbose(s):
    if VERBOSE or DEBUG:
        printStderr('[V] ' + str(s))

def debug(s):
    if DEBUG:
        printStderr('[D] ' + str(s))

def warn(s: str):
    printStderr('WARN: ' + str(s))



import types
import traceback
import utils
import inspect
from typing import Optional, Any
import os
import sys
from collections import deque

def tbToFrameList(tb: types.TracebackType) -> list[types.FrameType]:
    cur = tb
    res: list[types.FrameType] = []
    while cur:
        res.append(cur.tb_frame)
        cur = cur.tb_next
    return res

def isCallWithFramesRemoved(frame: types.FrameType):
    return frame.f_code.co_name == utils._call_with_frames_removed.__name__

def isCallWithNextFrameRemoved(frame: types.FrameType):
    return frame.f_code.co_name == utils._call_with_next_frame_removed.__name__

def isWyppFrame(frame: types.FrameType):
    modName = frame.f_globals.get("__name__") or '__wypp__'
    fname = frame.f_code.co_filename
    directDir = os.path.basename(os.path.dirname(fname))
    if '__wypp_runYourProgram' in frame.f_globals:
        return True
    for prefix in ['wypp', 'typeguard']:
        if modName == prefix or modName.startswith(prefix + '.') or directDir == prefix:
            return True
    return False

def isRunpyFrame(frame: types.FrameType) -> bool:
    f = frame.f_code.co_filename
    return f == '<frozen runpy>' or f.startswith('<frozen importlib.')

# Returns a StackSummary object. Filters the trackback by removing leading wypp or typeguard
# frames and by removing trailing frames behind _call_with_frames_removed.
# The first entry is the outermost frame, the last entry in the frame list is the frame
# where the exception happened.
def limitTraceback(frameList: list[types.FrameType],
                   extraFrames: list[inspect.FrameInfo],
                   filter: bool) -> traceback.StackSummary:
    if filter:
        # Step 1: remove all frames that appear after the first _call_with_frames_removed
        endIdx = len(frameList)
        for i in range(len(frameList)):
            if isCallWithFramesRemoved(frameList[i]):
                endIdx = i - 1
                break
        frameList = frameList[:endIdx]
        # Step 2: remove those frames directly after _call_with_next_frame_removed
        toRemove = []
        for i in range(len(frameList)):
            if isCallWithNextFrameRemoved(frameList[i]):
                toRemove.append(i)
        for i in reversed(toRemove):
            frameList = frameList[:i-1] + frameList[i+1:]
        # Step 3: remove leading wypp or typeguard frames
        frameList = utils.dropWhile(frameList, lambda f: isWyppFrame(f) or isRunpyFrame(f))
    frameList = frameList + [f.frame for f in extraFrames]
    frames = [(f, f.f_lineno) for f in frameList]
    return traceback.StackSummary.extract(frames)

def callerOutsideWypp() -> Optional[inspect.FrameInfo]:
    stack = inspect.stack()
    d = os.path.dirname(stack[0].filename)
    for f in stack:
        if not isWyppFrame(f.frame) and not d == os.path.dirname(f.filename):
            return f
    return None

class ReturnTracker:
    def __init__(self, entriesToKeep: int):
        self.__returnFrames = deque(maxlen=entriesToKeep)   # a ring buffer
    def __call__(self, frame: types.FrameType, event: str, arg: Any):
        # event is one of 'call', 'return', 'c_call', 'c_return', or 'c_exception'
        match event:
            case 'call':
                pass
            case 'return':
                # print(f'appending {frame} to return tracker')
                self.__returnFrames.append(frame) # overwrite oldest when full
            case 'c_call':
                pass
            case 'c_return':
                pass
            case 'c_exception':
                pass
    def getReturnFrame(self, idx: int) -> Optional[inspect.FrameInfo]:
        try:
            f = self.__returnFrames[idx]
        except IndexError:
            return None
        if f:
            tb = inspect.getframeinfo(f, context=1)
            fi = inspect.FrameInfo(f, tb.filename, tb.lineno, tb.function, tb.code_context, tb.index)
            del f
            return fi
        else:
            return None

# when using _call_with_next_frame_removed, we have to take the second-to-last
# return. Hence, we keep the two most recent returns byn setting entriesToKeep = 2.
def installProfileHook(entriesToKeep: int=2) -> ReturnTracker:
    obj = sys.getprofile()
    if isinstance(obj, ReturnTracker):
        return obj
    obj = ReturnTracker(entriesToKeep)
    sys.setprofile(obj)
    return obj

def getReturnTracker():
    obj = sys.getprofile()
    if isinstance(obj, ReturnTracker):
        return obj
    else:
        raise ValueError(f'No ReturnTracker set, must use installProfileHook before')



import argparse
import sys
import utils
from myLogging import *

def parseCmdlineArgs(argList):
    parser = argparse.ArgumentParser(description='Run Your Program!',
                        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--check-runnable', dest='checkRunnable', action='store_const',
                        const=True, default=False,
                        help='Abort with exit code 1 if loading the file raises errors')
    parser.add_argument('--check', dest='check', action='store_const',
                        const=True, default=False,
                        help='Abort with exit code 1 if there are test errors.')
    parser.add_argument('--verbose', dest='verbose', action='store_const',
                        const=True, default=False,
                        help='Be verbose')
    parser.add_argument('--debug', dest='debug', action='store_const',
                        const=True, default=False,
                        help='Enable debugging')
    parser.add_argument('--quiet', dest='quiet', action='store_const',
                        const=True, default=False, help='Be extra quiet')
    parser.add_argument('--lang', dest='lang',
                        type=str, help='Display error messages in this language (either en or de).')
    parser.add_argument('--no-clear', dest='noClear', action='store_const',
                        const=True, default=False, help='Do not clear the terminal')
    parser.add_argument('--test-file', dest='testFile',
                        type=str, help='Run additional tests contained in this file.')
    parser.add_argument('--extra-dir', dest='extraDirs', action='append', type=str,
                        help='Also typechecks files contained in the given directory.\n' \
                            'By default, only files in the same directory as the main file are\n' \
                            'checked.')
    parser.add_argument('--change-directory', dest='changeDir', action='store_const',
                        const=True, default=False,
                        help='Change to the directory of FILE before running')
    parser.add_argument('--interactive', dest='interactive', action='store_const',
                        const=True, default=False,
                        help='Run REPL after the programm has finished')
    parser.add_argument('--no-typechecking', dest='checkTypes', action='store_const',
                        const=False, default=True,
                        help='Do not check type annotations')
    parser.add_argument('--repl', action='extend', type=str, nargs='+', default=[], dest='repls',
                        help='Run repl tests in the file given')
    parser.add_argument('file', metavar='FILE',
                        help='The file to run', nargs='?')
    if argList is None:
        argList = sys.argv[1:]
    try:
        args, restArgs = parser.parse_known_args(argList)
    except SystemExit as ex:
        utils.die(ex.code)
    if args.file and not args.file.endswith('.py'):
        printStderr(f'ERROR: file {args.file} is not a python file')
        utils.die()
    return (args, restArgs)



try:
    from . import writeYourProgram as w
except (ImportError, ModuleNotFoundError):
    import writeYourProgram as w

import typing

# Exported names that are available for star imports (mostly in alphabetic order)
from typing import Any, Callable, Generator, Iterable, Iterator, Literal, Mapping, Optional, \
    Protocol, Sequence, Union
from dataclasses import dataclass

check = w.check
checkFail = w.checkFail
floatNegative = w.floatNegative
floatNonNegative = w.floatNonNegative
floatNonPositive = w.floatNonPositive
floatPositive = w.floatPositive
impossible = w.impossible
intNegative = w.intNegative
intNonNegative = w.intNonNegative
intNonPositive = w.intNonPositive
intPositive = w.intPositive
math = w.math
nat = w.nat

@typing.dataclass_transform()
def record(cls=None, mutable=False, globals={}, locals={}):
    return w.record(cls, mutable, globals, locals)

T = w.T
todo = w.todo
U = w.U
V = w.V

__all__ = [
    'Any',
    'Callable',
    'Generator',
    'Iterable',
    'Iterator',
    'Literal',
    'Mapping',
    'Optional',
    'Sequence',
    'Protocol',
    'Union',
    'check',
    'checkFail',
    'dataclass',
    'floatNegative',
    'floatNonNegative',
    'floatNonPositive',
    'floatPositive',
    'impossible',
    'intNegative',
    'intNonNegative',
    'intNonPositive',
    'intPositive',
    'math',
    'nat',
    'record',
    'T',
    'todo',
    'U',
    'V'
]

# Exported names not available for star imports (in alphabetic order)
Lock = w.Lock
LockFactory = w.LockFactory
initModule = w.initModule
printTestResults = w.printTestResults
resetTestCount = w.resetTestCount
deepEq = w.deepEq
wrapTypecheck = w.wrapTypecheck



import unittest
from utils import dropWhile

class TestUtils(unittest.TestCase):

    def test_dropWhile(self):
        self.assertEqual(dropWhile([1, 2, 3, 4, 5], lambda x: x < 4), [4, 5])
        self.assertEqual(dropWhile([], lambda x: x < 10), [])
        # Test where no elements satisfy the condition
        self.assertEqual(dropWhile([5, 6, 7, 8], lambda x: x < 3), [5, 6, 7, 8])
        # Test where all elements satisfy the condition
        self.assertEqual(dropWhile([1, 2, 3, 4], lambda x: x < 10), [])



import os
from typing import *
from contextlib import contextmanager

_projectDir: Optional[str] = None

@contextmanager
def projectDir(d: str):
    """Context manager to temporarily set the _relDir global variable."""
    global _projectDir
    old = _projectDir
    _projectDir = _normPath(d)
    sep = os.path.sep
    if _projectDir and _projectDir[-1] != sep:
        _projectDir = _projectDir + sep
    try:
        yield
    finally:
        _projectDir = old

def _normPath(s: str) -> str:
    return os.path.normpath(os.path.abspath(s))

def canonicalizePath(s: str) -> str:
    s = _normPath(s)
    if _projectDir and s.startswith(_projectDir):
        s = s[len(_projectDir):]
    return s



import unittest
from writeYourProgram import *
from typing import *

setDieOnCheckFailures(True)

ty = ForwardRef('Name')

@record
class Name:
    firstName: str
    lastName: str

@record
class Point:
    x: float
    y: float

PointOrFloat = Union[Point, float]

WithNone = Union[str, None]

class C:
    pass

def abcGenerator():
    yield("a")
    yield("b")
    yield("c")

class TestMisc(unittest.TestCase):

    def test_types(self):
        # just check that the types are defined
        Mapping
        dict
        dict[str, int]
        myName = Name("Stefan", "Wehr")
        list[Name] # just use it
        list[ty] # just use it
        list[ForwardRef('foo')] # just use
        list[PointOrFloat]

    def test_math(self):
        # just check that the functions are defined
        math.sqrt(2)
        math.sin(2)

    def assertTypeError(self, thunk, msg):
        try:
            thunk()
            self.fail("no type error")
        except TypeError as e:
            self.assertEqual(str(e), msg)

    def test_checkEq(self):
        check(1.0000000000000001, 1)

    def test_typechking(self):
       self.assertTypeError(lambda : ForwardRef(1),
            'Forward reference must be a string -- got 1')



import unittest
import parsecache
from parsecache import FunMatcher
import ast

class TestParseCache(unittest.TestCase):

    def test_parseCacheFunDef(self):
        a = parsecache.AST('tests/parsecacheTestData.py')
        assert(a is not None)
        defFoo = a.getFunDef(FunMatcher('foo', 6))
        assert(defFoo is not None)
        self.assertEqual(defFoo.lineno, 6)
        defBar = a.getFunDef(FunMatcher('bar'))
        assert(defBar is not None)
        self.assertEqual(defBar.lineno, 3)
        x = a.getFunDef(FunMatcher('spam')) # conflicting def
        self.assertIsNone(x)
        defSpam = a.getFunDef(FunMatcher('spam', 22))
        assert(defSpam is not None)
        self.assertEqual(defSpam.lineno, 22)
        match defSpam.body[1]:
            case ast.Return(ast.Constant(2)):
                pass
            case stmt:
                self.fail(f'Invalid first statement of spam: {stmt}')
        x = a.getFunDef(FunMatcher('egg'))
        self.assertIsNone(x)

    def test_parseCacheNestedFunDef(self):
        a = parsecache.AST('tests/parsecacheTestData.py')
        assert(a is not None)
        defFoo = a.getFunDef(FunMatcher('foo', 23))
        assert(defFoo is not None)
        self.assertEqual(defFoo.lineno, 23)

    def test_parseCacheMethod(self):
        a = parsecache.AST('tests/parsecacheTestData.py')
        assert(a is not None)
        defFoo = a.getMethodDef('D', FunMatcher('foo', 16))
        assert(defFoo is not None)
        self.assertEqual(defFoo.lineno, 16)

    def test_parseCacheRecordAttr(self):
        a = parsecache.AST('tests/parsecacheTestData.py')
        assert(a is not None)

        # Test getting attributes from class C
        attrX_C = a.getRecordAttr('C', 'x')
        assert(attrX_C is not None)
        self.assertEqual(attrX_C.lineno, 10)

        attrY_C = a.getRecordAttr('C', 'y')
        assert(attrY_C is not None)
        self.assertEqual(attrY_C.lineno, 11)

        # Test getting attributes from class D
        attrX_D = a.getRecordAttr('D', 'x')
        assert(attrX_D is not None)
        self.assertEqual(attrX_D.lineno, 14)

        attrZ_D = a.getRecordAttr('D', 'z')
        assert(attrZ_D is not None)
        self.assertEqual(attrZ_D.lineno, 15)

        # Test non-existent attribute
        nonExistent = a.getRecordAttr('C', 'z')
        self.assertIsNone(nonExistent)

        # Test non-existent class
        nonExistentClass = a.getRecordAttr('NonExistentClass', 'x')
        self.assertIsNone(nonExistentClass)






import sys
import traceback
import re
from dataclasses import dataclass

# local imports
from constants import *
import stacktrace
import paths
from myLogging import *
import errors
import utils

_tbPattern = re.compile(r'(\s*File\s+")([^"]+)(".*)')
def _rewriteFilenameInTracebackLine(s: str) -> str:
    # Match the pattern: File "filename", line number, in function
    match = re.match(_tbPattern, s)
    if match:
        prefix = match.group(1)
        filename = match.group(2)
        suffix = match.group(3)
        canonicalized = paths.canonicalizePath(filename)
        return prefix + canonicalized + suffix
    else:
        return s

def _rewriteFilenameInTraceback(s: str) -> str:
    lines = s.split('\n')
    result = []
    for l in lines:
        result.append(_rewriteFilenameInTracebackLine(l))
    return '\n'.join(result)

def handleCurrentException(exit=True, removeFirstTb=False, file=sys.stderr):
    (etype, val, tb) = sys.exc_info()
    if isinstance(val, SystemExit):
        utils.die(val.code)
    isWyppError = isinstance(val, errors.WyppError)
    if isWyppError:
        extra = val.extraFrames
    else:
        extra = []
    frameList = (stacktrace.tbToFrameList(tb) if tb is not None else [])
    if frameList and removeFirstTb:
        frameList = frameList[1:]
    isBug = not isWyppError and not isinstance(val, SyntaxError) and \
        len(frameList) > 0 and stacktrace.isWyppFrame(frameList[-1])
    stackSummary = stacktrace.limitTraceback(frameList, extra, not isBug and not isDebug())
    header = False
    for x in stackSummary.format():
        if not header:
            file.write('Traceback (most recent call last):\n')
            header = True
        file.write(_rewriteFilenameInTraceback(x))
    if isWyppError:
        s = str(val)
        if s and s[0] != '\n':
            file.write('\n')
        file.write(s)
        file.write('\n')
    else:
        for x in traceback.format_exception_only(etype, val):
            file.write(x)
    if isBug:
        file.write(f'BUG: the error above is most likely a bug in WYPP!')
    if exit:
        utils.die(1)



import unittest
import sys
import types
import stacktrace
from stacktraceTestData import *
import os

class TestReturnTracker(unittest.TestCase):

    def setUp(self):
        self.original_profile = sys.getprofile()

    def tearDown(self):
        sys.setprofile(self.original_profile)

    def assertReturnFrame(self, tracker: stacktrace.ReturnTracker, line: int):
        frame = tracker.getReturnFrame(0)
        assert(frame is not None)
        self.assertEqual(os.path.basename(frame.filename), 'stacktraceTestData.py')
        self.assertEqual(frame.lineno, line)

    def test_returnTracker1(self):
        tracker = stacktrace.installProfileHook(1)
        f1()
        self.assertReturnFrame(tracker, f1ReturnLine)

    def test_returnTracker2(self):
        tracker = stacktrace.installProfileHook(1)
        f2()
        self.assertReturnFrame(tracker, f2ReturnLine)

    def test_returnTracker3(self):
        tracker = stacktrace.installProfileHook(1)
        f3()
        self.assertReturnFrame(tracker, f3ReturnLine)



import unittest
from writeYourProgram import *
import sys
import traceback
import dataclasses
import stacktrace
from typing import Literal

initModule()

@record
class Point:
    x: float
    y: float

@record
class Square:
    center: Point
    size: float

@record
class Name:
    firstName: str
    lastName: str

@record(mutable=True)
class Box:
    x: int

class TestRecords(unittest.TestCase):

    def setUp(self):
        setDieOnCheckFailures(True)
        self.original_profile = sys.getprofile()
        stacktrace.installProfileHook()

    def tearDown(self):
        # Restore original profile function
        sys.setprofile(self.original_profile)
        setDieOnCheckFailures(False)

    def test_create(self):
        p1 = Point(1, 2)
        p2 = Point(3, 4)
        self.assertEqual(p1.x, 1)
        self.assertEqual(p1.y, 2)
        self.assertEqual(p2.x, 3)
        self.assertEqual(p2.y, 4)
        square = Square(p1, 5)
        self.assertEqual(square.center, p1)
        self.assertEqual(square.size, 5)
        list[Point] # just use it
        list[Literal['foo', 'bar']] # just use it

    def test_createErrorArity(self):
        pass # FIXME

    def test_createErrorTypes(self):
        pass # FIXME

    def test_toString(self):
        p1 = Point(1, 2)
        self.assertEqual(str(p1), 'Point(x=1, y=2)')
        square = Square(p1, 5)
        self.assertEqual(str(square), 'Square(center=Point(x=1, y=2), size=5)')
        name = Name("Stefan", "Wehr")
        self.assertEqual(str(name), "Name(firstName='Stefan', lastName='Wehr')")

    def test_eq(self):
        p1 = Point(1, 2)
        p2 = Point(1, 4)
        p3 = Point(1, 2)
        self.assertEqual(p1, p1)
        self.assertEqual(p1, p3)
        self.assertNotEqual(p1, p2)
        s1 = Square(p1, 5)
        s2 = Square(p3, 5)
        s3 = Square(p3, 6)
        s4 = Square(p2, 5)
        self.assertEqual(s1, s1)
        self.assertEqual(s1, s2)
        self.assertNotEqual(s1, s3)
        self.assertNotEqual(s1, s4)

    def test_eq2(self):
        p1 = Point(0.3, 2)
        p2 = Point(0.1 + 0.1 + 0.1, 2)
        # general equality must not use special logic for floats, otherwise == is not transitive
        self.assertEqual(p1 == p2, False)
        # check should use special logic for floats
        check(p1, p2)

    def test_hash(self):
        p1 = Point(1, 2)
        p2 = Point(1, 4)
        p3 = Point(1, 2)
        self.assertEqual(hash(p1), hash(p3))
        self.assertNotEqual(hash(p1), hash(p2))

    def test_unknownSelector(self):
        p1 = Point(1, 2)
        excMsg = None
        # p1.bar
        try:
            p1.foo
        except AttributeError:
            excMsg = traceback.format_exc()
        if excMsg is None:
            self.fail("Expected an AttributeError")
        self.assertTrue("'Point' object has no attribute 'foo'" in excMsg)

    def test_immutable(self):
        p1 = Point(1, 2)
        try:
            p1.x = 5
            self.fail('Expected FrozenInstanceError')
        except dataclasses.FrozenInstanceError:
            pass

    def test_mutable(self):
        b = Box(4)
        self.assertEqual(b.x, 4)
        b.x = 5
        self.assertEqual(b.x, 5)

    def test_addField(self):
        b = Box(5)
        try:
            b.foobar = 'foobar'
            self.fail(f"Expected AttributeError, b={b}")
        except AttributeError:
            pass
        p1 = Point(1, 2)
        try:
            p1.z = 5
            self.fail('Expected FrozenInstanceError')
        except dataclasses.FrozenInstanceError:
            pass



import unittest
from typing import *
from renderTy import renderTy

class MyClass:
    pass

class TestRenderTy(unittest.TestCase):

    def test_basics(self):
        self.assertEqual(renderTy(int), 'int')
        self.assertEqual(renderTy(list[str]), 'list[str]')
        self.assertEqual(renderTy(dict[str, list[int]]), 'dict[str, list[int]]')
        self.assertEqual(renderTy(Optional[dict[str, int]]), 'Optional[dict[str, int]]')
        self.assertEqual(renderTy(tuple[int, ...]), 'tuple[int, ...]')
        self.assertEqual(renderTy(Union[int, str, bool]), 'Union[int, str, bool]')
        self.assertEqual(renderTy(Callable[[int, str, list[float]], bool]),
                         'Callable[[int, str, list[float]], bool]')

    def test_advanced_types(self):
        # Nested generics
        self.assertEqual(renderTy(dict[str, list[tuple[int, str]]]), 'dict[str, list[tuple[int, str]]]')
        self.assertEqual(renderTy(list[dict[str, Optional[int]]]), 'list[dict[str, Optional[int]]]')

        # Complex Union types
        self.assertEqual(renderTy(Union[int, str, None]), 'Union[int, str, None]')
        self.assertEqual(renderTy(Union[list[int], dict[str, int], tuple[int, ...]]),
                         'Union[list[int], dict[str, int], tuple[int, ...]]')
        self.assertEqual(renderTy(int | str), 'Union[int, str]')
        self.assertEqual(renderTy(list[int] | dict[str, int] | None),
                         'Union[list[int], dict[str, int], None]')

        # Callable with complex signatures
        self.assertEqual(renderTy(Callable[[dict[str, int], Optional[list[str]]], Union[int, str]]),
                         'Callable[[dict[str, int], Optional[list[str]]], Union[int, str]]')
        self.assertEqual(renderTy(Callable[[], None]), 'Callable[[], None]')

        # Literal types
        self.assertEqual(renderTy(Literal['red', 'green', 'blue']), "Literal['red', 'green', 'blue']")
        self.assertEqual(renderTy(Literal[1, 2, 3]), 'Literal[1, 2, 3]')

    def test_special_forms(self):
        # Any and NoReturn
        self.assertEqual(renderTy(Any), 'Any')
        self.assertEqual(renderTy(NoReturn), 'NoReturn')

        # Forward references (if supported)
        self.assertEqual(renderTy(ForwardRef('SomeClass')), 'SomeClass')


    def test_edge_cases(self):
        # Empty tuple
        self.assertEqual(renderTy(tuple[()]), 'tuple[()]')

        # Single element tuple
        self.assertEqual(renderTy(tuple[int]), 'tuple[int]')

        # Very nested types
        self.assertEqual(renderTy(dict[str, dict[str, dict[str, int]]]),
                         'dict[str, dict[str, dict[str, int]]]')

        # Union with single type
        self.assertEqual(renderTy(Union[int]), 'int')

        # Complex Callable with nested types
        self.assertEqual(renderTy(Callable[[Callable[[int], str], list[int]], dict[str, Any]]),
                         'Callable[[Callable[[int], str], list[int]], dict[str, Any]]')

    def test_custom_classes(self):
        # Custom class types

        self.assertEqual(renderTy(MyClass), 'tests.test_renderTy.MyClass')
        self.assertEqual(renderTy(list[MyClass]), 'list[tests.test_renderTy.MyClass]')
        self.assertEqual(renderTy(Optional[MyClass]), 'Optional[tests.test_renderTy.MyClass]')




from __future__ import annotations
from typing import *
import abc
import inspect
import location
import i18n
from renderTy import renderTy

class WyppError(abc.ABC):
    def __init__(self, extraFrames: list[inspect.FrameInfo] = []):
        self.extraFrames = extraFrames[:]

def renderLoc(loc: location.Loc) -> str:
    s = '\n'.join([l.highlight() for l in location.highlightedLines(loc)])
    return s.rstrip()

def renderStr(s: str, quote: str):
    x = repr(s)
    return quote + x[1:-1] + quote

def renderGiven(givenValue: Any, givenLoc: Optional[location.Loc]) -> str:
    if type(givenValue) == str and givenLoc is not None:
        code = renderLoc(givenLoc)
        single = renderStr(givenValue, "'")
        double = renderStr(givenValue, '"')
        if double in code and single not in code:
            return double
        elif '"' in code and "'" not in code:
            return double
        else:
            return single
    else:
        return repr(givenValue)

def isParameterizedType(t: Any) -> bool:
    """True for any parameterized typing construct (incl. Union, Annotated, etc.)."""
    return get_origin(t) is not None and len(get_args(t)) > 0

def shouldReportTyMismatch(expected: Any, given: Any) -> bool:
    if isParameterizedType(expected) and get_origin(expected) == given:
        # don't report a mismatch because givenTy will not be properly parameterized.
        # Example: resultTy is `list[int]` and givenValue is [1, "blub"]. Then
        # givenTy will be just `list`
        return False
    else:
        return True

def rewriteInvalidType(s: str) -> Optional[str]:
    prefixes = ['Union', 'Literal', 'Optional', 'list', 'dict', 'tuple', 'set']
    for p in prefixes:
        if s.startswith(f'{p}(') and s.endswith(')'):
            args = s[len(p)+1:-1]
            return f'{p}[{args}]'

class WyppTypeError(TypeError, WyppError):

    def __init__(self, msg: str, extraFrames: list[inspect.FrameInfo] = []):
        WyppError.__init__(self, extraFrames)
        self.msg = msg
        self.add_note(msg)

    def __str__(self):
        return f'WyppTypeError: {self.msg}'

    @staticmethod
    def invalidType(ty: Any, loc: Optional[location.Loc]) -> WyppTypeError:
        lines = []
        tyStr = renderTy(ty)
        lines.append(i18n.invalidTy(tyStr))
        lines.append('')
        rew = rewriteInvalidType(tyStr)
        if rew:
            lines.append(i18n.didYouMean(rew))
            lines.append('')
        if loc is not None:
            lines.append(f'## {i18n.tr("File")} {loc.filename}')
            lines.append(f'## {i18n.tr("Type declared in line")} {loc.startLine}:\n')
            lines.append(renderLoc(loc))
        raise WyppTypeError('\n'.join(lines))

    @staticmethod
    def unknownKeywordArgument(callableName: location.CallableName, callLoc: Optional[location.Loc], name: str) -> WyppTypeError:
        lines = []
        lines.append(i18n.tr('unknown keyword argument'))
        lines.append('')
        lines.append(i18n.unknownKeywordArgument(callableName, name))
        if callLoc and (callLocR := renderLoc(callLoc)):
            lines.append('')
            lines.append(f'## {i18n.tr("File")} {callLoc.filename}')
            lines.append(f'## {i18n.tr("Problematic call in line")} {callLoc.startLine}:\n')
            lines.append(callLocR)
        raise WyppTypeError('\n'.join(lines))

    @staticmethod
    def invalidRecordAnnotation(loc: Optional[location.Loc]) -> WyppTypeError:
        lines = []
        lines.append(i18n.tr('invalid record definition'))
        lines.append('')
        if loc and (locR := renderLoc(loc)):
            lines.append(f'## {i18n.tr("File")} {loc.filename}')
            lines.append(f'## {i18n.tr("Line")} {loc.startLine}:\n')
            lines.append(locR)
        raise WyppTypeError('\n'.join(lines))

    @staticmethod
    def resultError(callableName: location.CallableName, resultTypeLoc: Optional[location.Loc], resultTy: Any,
                    returnLoc: Optional[location.Loc], givenValue: Any,
                    callLoc: Optional[location.Loc],
                    extraFrames: list[inspect.FrameInfo]) -> WyppTypeError:
        lines = []
        if givenValue is None:
            lines.append(i18n.tr('no result returned'))
        else:
            lines.append(renderGiven(givenValue, returnLoc))
        lines.append('')
        if resultTy is None:
            # no result type expected but given
            lines.append(i18n.expectingNoReturn(callableName))
            lines.append(i18n.wrongReturnValue(renderTy(type(givenValue))))
        elif givenValue is None:
            # result type expected but none given
            lines.append(i18n.expectingReturnOfType(callableName, renderTy(resultTy)))
            lines.append(i18n.noReturnValue())
        else:
            # result type expected but different type given
            lines.append(i18n.expectingReturnOfType(callableName, renderTy(resultTy)))
            givenTy = type(givenValue)
            if shouldReportTyMismatch(resultTy, givenTy):
                lines.append(i18n.wrongReturnValue(renderTy(givenTy)))
        printedFileName = None
        if resultTypeLoc and (resultTypeLocR := renderLoc(resultTypeLoc)):
            lines.append('')
            lines.append(f'## {i18n.tr("File")} {resultTypeLoc.filename}')
            printedFileName = resultTypeLoc.filename
            lines.append(f'## {i18n.tr("Result type declared in line")} {resultTypeLoc.startLine}:\n')
            lines.append(resultTypeLocR)
        if givenValue is not None and returnLoc and (returnLocR := renderLoc(returnLoc)):
            lines.append('')
            if printedFileName != returnLoc.filename:
                lines.append(f'## {i18n.tr("File")} {returnLoc.filename}')
                printedFileName = returnLoc.filename
            lines.append(f'## {i18n.tr("Problematic return in line")} {returnLoc.startLine}:\n')
            lines.append(returnLocR)
        if callLoc and (callLocR := renderLoc(callLoc)):
            lines.append('')
            if printedFileName != callLoc.filename:
                lines.append(f'## {i18n.tr("File")} {callLoc.filename}')
            if givenValue is None:
                lines.append(f'## {i18n.unexpectedNoReturn(callLoc.startLine)}\n')
            else:
                lines.append(f'## {i18n.unexpectedReturn(callLoc.startLine)}\n')
            lines.append(callLocR)
        raise WyppTypeError('\n'.join(lines), extraFrames)

    @staticmethod
    def argumentError(callableName: location.CallableName, paramName: str, paramIndex: Optional[int],
                      paramLoc: Optional[location.Loc],
                      paramTy: Any, givenValue: Any, givenLoc: Optional[location.Loc]) -> WyppTypeError:
        lines = []
        givenStr = renderGiven(givenValue, givenLoc)
        lines.append(givenStr)
        lines.append('')
        if paramIndex is not None:
            lines.append(i18n.expectingArgumentOfTy(callableName, renderTy(paramTy), paramIndex + 1))
        else:
            lines.append(i18n.expectingArgumentOfTy(callableName, renderTy(paramTy), paramName))
        if shouldReportTyMismatch(paramTy, type(givenValue)):
            lines.append(i18n.realArgumentTy(renderTy(type(givenValue))))
        if givenLoc and (givenLocR := renderLoc(givenLoc)):
            lines.append('')
            lines.append(f'## {i18n.tr("File")} {givenLoc.filename}')
            lines.append(f'## {i18n.tr("Problematic call in line")} {givenLoc.startLine}:\n')
            lines.append(givenLocR)
        if paramLoc and (paramLocR := renderLoc(paramLoc)):
            lines.append('')
            if not givenLoc or paramLoc.filename != givenLoc.filename:
                lines.append(f'## {i18n.tr("File")} {paramLoc.filename}')
            lines.append(f'## {i18n.tr("Type declared in line")} {paramLoc.startLine}:\n')
            lines.append(paramLocR)
        raise WyppTypeError('\n'.join(lines))

    @staticmethod
    def defaultError(callableName: location.CallableName, paramName: str, paramLoc: Optional[location.Loc],
                      paramTy: Any, givenValue: Any) -> WyppTypeError:
        lines = []
        givenStr = renderGiven(givenValue, paramLoc)
        lines.append(givenStr)
        lines.append('')
        lines.append(i18n.expectingDefaultValueOfTy(callableName, renderTy(paramTy), paramName))
        if shouldReportTyMismatch(paramTy, type(givenValue)):
            lines.append(i18n.realDefaultValueTy(renderTy(type(givenValue))))
        if paramLoc and (paramLocR := renderLoc(paramLoc)):
            lines.append('')
            lines.append(f'## {i18n.tr("File")} {paramLoc.filename}')
            lines.append(f'## {i18n.tr("Parameter declared in line")} {paramLoc.startLine}:\n')
            lines.append(paramLocR)
        raise WyppTypeError('\n'.join(lines))

    @staticmethod
    def argCountMismatch(callableName: location.CallableName, callLoc: Optional[location.Loc],
                         numParams: int, numMandatoryParams: int, numArgs: int) -> WyppTypeError:
        lines = []
        lines.append(i18n.tr('argument count mismatch'))
        lines.append('')
        if numArgs < numMandatoryParams:
            if numParams == numMandatoryParams:
                lines.append(i18n.argCountExact(callableName, numParams))
            else:
                lines.append(i18n.argCountMin(callableName, numMandatoryParams))
        else:
            if numParams == numMandatoryParams:
                lines.append(i18n.argCountExact(callableName, numParams))
            else:
                lines.append(i18n.argCountMax(callableName, numParams))
        lines.append(i18n.tr('Given: ') + i18n.argCount(numArgs))
        if callLoc and (callLocR := renderLoc(callLoc)):
            lines.append('')
            lines.append(f'## {i18n.tr("File")} {callLoc.filename}')
            lines.append(f'## {i18n.tr("Call in line")} {callLoc.startLine}:\n')
            lines.append(callLocR)
        raise WyppTypeError('\n'.join(lines))

    @staticmethod
    def partialAnnotationError(callableName: location.CallableName, paramName: str, paramLoc: Optional[location.Loc]) -> WyppTypeError:
        lines = []
        lines.append(i18n.expectingTypeAnnotation(callableName, paramName))
        if paramLoc and (paramLocR := renderLoc(paramLoc)):
            lines.append('')
            lines.append(f'## {i18n.tr("File")} {paramLoc.filename}')
            lines.append(f'## {i18n.tr("Parameter declared in line")} {paramLoc.startLine}:\n')
            lines.append(paramLocR)
        raise WyppTypeError('\n'.join(lines))

    @staticmethod
    def noTypeAnnotationForRecordAttribute(attrName: str, recordName: str) -> WyppTypeError:
        return WyppTypeError(i18n.noTypeAnnotationForAttribute(attrName, recordName))

    @staticmethod
    def recordAssignError(recordName: str,
                          attrName: str,
                          attrTy: Any,
                          attrLoc: Optional[location.Loc],
                          setterValue: Any,
                          setterLoc: Optional[location.Loc]) -> WyppTypeError:
        lines = []
        givenStr = renderGiven(setterValue, setterLoc)
        lines.append(givenStr)
        lines.append('')
        lines.append(i18n.recordAttrDeclTy(recordName, attrName, renderTy(attrTy)))
        if shouldReportTyMismatch(attrTy, type(setterValue)):
            lines.append(i18n.realSetAttrTy(renderTy(type(setterValue))))
        if setterLoc and (setterLocR := renderLoc(setterLoc)):
            lines.append('')
            lines.append(f'## {i18n.tr("File")} {setterLoc.filename}')
            lines.append(f'## {i18n.tr("Problematic assignment in line")} {setterLoc.startLine}:\n')
            lines.append(setterLocR)
        if attrLoc and (attrLocR := renderLoc(attrLoc)):
            lines.append('')
            if not setterLoc or setterLoc.filename != attrLoc.filename:
                lines.append(f'## {i18n.tr("File")} {attrLoc.filename}')
            lines.append(f'## {i18n.tr("Type declared in line")} {attrLoc.startLine}:\n')
            lines.append(attrLocR)
        raise WyppTypeError('\n'.join(lines))

class WyppAttributeError(AttributeError, WyppError):
    def __init__(self, msg: str, extraFrames: list[inspect.FrameInfo] = []):
        WyppError.__init__(self, extraFrames)
        self.msg = msg
        self.add_note(msg)

    @staticmethod
    def unknownAttr(clsName: str, attrName: str) -> WyppAttributeError:
        return WyppAttributeError(i18n.tr('Unknown attribute {attrName} for record {clsName}',
                                          clsName=clsName, attrName=attrName))

class TodoError(Exception, WyppError):
    def __init__(self, msg: str, extraFrames: list[inspect.FrameInfo] = []):
        WyppError.__init__(self, extraFrames)
        self.msg = msg
        self.add_note(msg)


class ImpossibleError(Exception, WyppError):
    def __init__(self, msg: str, extraFrames: list[inspect.FrameInfo] = []):
        WyppError.__init__(self, extraFrames)
        self.msg = msg
        self.add_note(msg)





import sys
import os
import code
from dataclasses import dataclass

# local imports
from constants import *
from myLogging import *
from exceptionHandler import handleCurrentException
import i18n

def prepareInteractive(reset=True):
    print()
    if reset:
        if os.name == 'nt':
            # clear the terminal
            os.system('cls')
        else:
            # On linux & mac use ANSI Sequence for this
            print('\033[2J\033[H')


HISTORY_SIZE = 1000

def getHistoryFilePath():
    envVar = 'HOME'
    if os.name == 'nt':
        envVar = 'USERPROFILE'
    d = os.getenv(envVar, None)
    if d:
        return os.path.join(d, ".wypp_history")
    else:
        return None

class TypecheckedInteractiveConsole(code.InteractiveConsole):
    def showtraceback(self) -> None:
        handleCurrentException(exit=False, removeFirstTb=True, file=sys.stdout)

def enterInteractive(userDefs: dict, checkTypes: bool, loadingFailed: bool):
    for k, v in userDefs.items():
        globals()[k] = v
    print()
    if checkTypes:
        consoleClass = TypecheckedInteractiveConsole
    else:
        consoleClass = code.InteractiveConsole
    historyFile = getHistoryFilePath()
    try:
        import readline
        readline.parse_and_bind('tab: complete')
        if historyFile and os.path.exists(historyFile):
            readline.read_history_file(historyFile)
    except:
        pass
    try:
        consoleClass(locals=userDefs).interact(banner="", exitmsg='')
    finally:
        if readline and historyFile:
            readline.set_history_length(HISTORY_SIZE)
            readline.write_history_file(historyFile)



import ast
import os
import linecache
from dataclasses import dataclass
from typing import *

def _firstLineOfFun(node: ast.FunctionDef | ast.AsyncFunctionDef) -> int:
    allLinenos = [node.lineno]
    for e in node.decorator_list:
        allLinenos.append(e.lineno)
    return min(allLinenos)

@dataclass(frozen=True)
class FunMatcher:
    name: str
    firstlineno: Optional[int] = None

    def matches(self, node: ast.FunctionDef | ast.AsyncFunctionDef) -> bool:
        if node.name == self.name:
            if self.firstlineno is not None:
                return self.firstlineno == _firstLineOfFun(node)
            else:
                return True
        else:
            return False

class AST:

    def __init__(self, filename: str):
        """
        Do not instantiate, use getAST
        """
        self.__filename = filename
        self.__tree = None
        self.__treeResolved = False
        self.__cache = {}

    def __ensureTree(self):
        if self.__treeResolved:
            return
        lines = linecache.getlines(self.__filename)
        try:
            tree = ast.parse(''.join(lines))
        except Exception:
            tree = None
        self.__tree = tree
        self.__treeResolved = True

    def _collectFuns(self, nodes: list[ast.stmt], m: FunMatcher, recursive: bool,
                     results: list[ast.FunctionDef | ast.AsyncFunctionDef]):
        for node in nodes:
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                if m.matches(node):
                    results.append(node)
                if recursive:
                    self._collectFuns(node.body, m, True, results)

    def _findFun(self, nodes: list[ast.stmt], m: FunMatcher, recursive: bool) -> ast.FunctionDef | ast.AsyncFunctionDef | None:
        """
        Finds the function from the list of statements that matches m. None is returned
        if no such function is found or if multiple matches are found.
        """
        results = []
        self._collectFuns(nodes, m, recursive, results)
        if len(results) == 1:
            return results[0]
        else:
            return None

    def getFunDef(self, m: FunMatcher) -> ast.FunctionDef | ast.AsyncFunctionDef | None:
        if m in self.__cache:
            return self.__cache[m]
        self.__ensureTree()
        if self.__tree:
            resultNode = self._findFun(self.__tree.body, m, True)
        self.__cache[m] = resultNode
        return resultNode

    def _findClass(self, clsName: str) -> ast.ClassDef | None:
        """
        Finds the toplevel class with the given name. Returns None if no such class or
        multiple classes are found.
        """
        targetClss = []
        if self.__tree:
            for node in self.__tree.body:
                if isinstance(node, ast.ClassDef) and node.name == clsName:
                    targetClss.append(node)
        if len(targetClss) == 1:
            return targetClss[0]
        else:
            return None

    def getRecordAttr(self, clsName: str, attrName: str) -> ast.AnnAssign | None:
        """
        Finds the attribute of the format `A: T` or `A: T = ...` with name attrName
        in a toplevel class with name clsName. Returns None if no such class/attribute
        comnbination or mutiples are found.
        """
        key = (clsName, attrName)
        if key in self.__cache:
            return self.__cache[key]
        self.__ensureTree()
        # Step 1: find the class
        targetCls = self._findClass(clsName)
        # Step 2: look for an AnnAssign like:   A: T = ...
        annNodes = []
        if targetCls:
            for stmt in targetCls.body:
                if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
                    if stmt.target.id == attrName:
                        annNodes.append(stmt)
        annNode = None
        if len(annNodes) == 1:
            annNode = annNodes[0]
        self.__cache[key] = annNode
        return annNode

    def getMethodDef(self, clsName: str, m: FunMatcher) -> ast.FunctionDef | ast.AsyncFunctionDef | None:
        """
        Finds the method in class clsName that matches m. None is returned if no such method
        or multiple methods are found.
        """
        key = (clsName, m)
        if key in self.__cache:
            return self.__cache[key]
        self.__ensureTree()
        # Step 1: find the class
        targetCls = self._findClass(clsName)
        # Step 2: look for the method
        funNode = None
        if targetCls:
            funNode = self._findFun(targetCls.body, m, False)
        self.__cache[key] = funNode
        # print(f'clsName={clsName}, m={m}, funNode={funNode}')
        return funNode

_cache: dict[str, AST] = {}

def getAST(filename: str) -> AST:
    filename = os.path.normpath(filename)
    if filename not in _cache:
        a = AST(filename)
        _cache[filename] = a
        return a
    else:
        return _cache[filename]



# Wrapper module for typeguard. Do not import typeguard directly but always via myTypeguard
from __future__ import annotations
from typing import *
from dataclasses import dataclass
from myLogging import *

# We externally adjust the PYTHONPATH so that the typeguard module can be resolved
import typeguard # type: ignore

@dataclass(frozen=True)
class Namespaces:
    globals: dict
    locals: dict
    @staticmethod
    def empty() -> Namespaces:
        return Namespaces({}, {})

@dataclass(frozen=True)
class MatchesTyFailure:
    # This is potentially a failure because of an invalid type
    exception: Exception
    ty: Any

type MatchesTyResult = bool | MatchesTyFailure

def matchesTy(a: Any, ty: Any, ns: Namespaces) -> MatchesTyResult:
    try:
        typeguard.check_type(a,
                             ty,
                             collection_check_strategy=typeguard.CollectionCheckStrategy.ALL_ITEMS,
                             ns=(ns.globals, ns.locals))
        return True
    except typeguard.TypeCheckError as e:
        return False
    except Exception as e:
        debug(f'Exception when checking type, ns={ns}: {e}')
        return MatchesTyFailure(e, ty)

def getTypeName(t: Any) -> str:
    res: str = typeguard._utils.get_type_name(t, ['__wypp__'])
    wyppPrefixes = ['wypp.writeYourProgram.']
    for p in wyppPrefixes:
        if res.startswith(p):
            return 'wypp.' + res[len(p):]
    return res



import os
import locale
from _collections_abc import MutableMapping

def _langFromEnv(env: MutableMapping) -> str | None:
    # 1) GNU LANGUAGE: colon-separated fallbacks (e.g., "de:en_US:en")
    lng = env.get("LANGUAGE")
    if lng:
        for part in lng.split(":"):
            part = part.strip()
            if part:
                return part

    # 2) POSIX locale vars in order of precedence
    for var in ("LC_ALL", "LC_MESSAGES", "LANG"):
        val = env.get(var)
        if val and val not in ("C", "POSIX"):
            return val

    # 3) locale module fallback (works cross-platform-ish)
    try:
        locale.setlocale(locale.LC_CTYPE, "")  # load user default
    except locale.Error:
        pass
    lang, _enc = locale.getlocale()  # e.g. "de_DE"
    return lang

def _normLang(tag: str) -> str:
    # "de-DE.UTF-8" -> "de_DE"
    tag = tag.replace("-", "_")
    if "." in tag:
        tag = tag.split(".", 1)[0]
    return tag

def pickLanguage[T: str](supported: list[T], default: T) -> T:
    """Return best match like 'de' or 'de_DE' from supported codes."""
    (raw, _) = locale.getlocale()
    if not raw:
        raw = _langFromEnv(os.environ)
        if not raw:
            return default
    want = _normLang(raw)
    # exact match first
    for s in supported:
        if _normLang(s).lower() == want.lower():
            return s
    # fallback to language-only match (de_DE -> de)
    wantBase = want.split("_")[0].lower()
    for s in supported:
        if _normLang(s).split('_')[0].lower() == wantBase:
            return s
    return default



import os
import json
import subprocess
from dataclasses import dataclass

# local imports
from constants import *
from myLogging import *
import utils

def readGitVersion():
    thisDir = os.path.basename(SOURCE_DIR)
    baseDir = os.path.join(SOURCE_DIR, '..', '..')
    if thisDir == 'src' and os.path.isdir(os.path.join(baseDir, '.git')):
        try:
            h = subprocess.check_output(['git', '-C', baseDir, 'rev-parse', '--short', 'HEAD'],
                encoding='UTF-8').strip()
            changes = subprocess.check_output(
                    ['git', '-C', baseDir, 'status', '--porcelain', '--untracked-files=no'],
                    encoding='UTF-8').strip()
            if changes:
                return f'git-{h}-dirty'
            else:
                return f'git-{h}'
        except subprocess.CalledProcessError:
            return 'git-?'
    else:
        return None

def readVersion():
    version = readGitVersion()
    if version is not None:
        return version
    try:
        content = utils.readFile(os.path.join(SOURCE_DIR, '..', '..', 'package.json'))
        d = json.loads(content)
        version = d['version']
    except:
        pass
    return version



import sys
import runner

# Use this file if you want to the some test file under the debugger
if __name__ == '__main__':
    sys.argv = ['python', 'test-data/testIndexError.py']
    runner.main(globals())



from __future__ import annotations
from typing import *
from dataclasses import dataclass
import inspect
import linecache
import dis
import ast
import ansi
import utils
import myLogging
import sys
import abc
import parsecache
from parsecache import FunMatcher
import paths
import tokenize
import os

@dataclass
class EncodedBytes:
    bytes: bytes
    encoding: str
    def __len__(self):
        return len(self.bytes)
    def countLeadingSpaces(self) -> int:
        return len(self.bytes) - len(self.bytes.lstrip())
    def decoded(self) -> str:
        return self.bytes.decode(self.encoding, errors='replace')
    @overload
    def __getitem__(self, key: int) -> int: ...
    @overload
    def __getitem__(self, key: slice) -> str: ...
    def __getitem__(self, key: int | slice) -> int | str:
        if isinstance(key, int):
            return self.bytes[key]
        else:
            b = self.bytes[key]
            return b.decode(self.encoding, errors='replace')

@dataclass
class EncodedByteLines:
    bytes: list[bytes]
    encoding: str

_cache: dict[str, EncodedByteLines] = {}
def getline(filename, lineno):
    """
    Returns a line of some source file as a bytearray. We use byte arrays because
    location offsets are byte offsets.
    """
    p = os.path.normpath(os.path.abspath(filename))
    if p in _cache:
        lines = _cache[p]
    else:
        try:
            with open(filename, 'rb') as f:
                byteLines = f.readlines()
        except Exception:
            byteLines = []
        i = 0
        def nextLine() -> bytes:
            nonlocal i
            if i < len(byteLines):
                x = byteLines[i]
                i = i + 1
                return x
            else:
                return b''
        encoding, _ = tokenize.detect_encoding(nextLine)
        lines = EncodedByteLines(byteLines, encoding)
    if 1 <= lineno <= len(lines.bytes):
        x = lines.bytes[lineno - 1].rstrip(b'\n')
    else:
        x = b''
    return EncodedBytes(x, encoding)

@dataclass
class Loc:
    filename: str
    startLine: int
    startCol: Optional[int]
    endLine: Optional[int]
    endCol: Optional[int]

    def __post_init__(self):
        self.filename = paths.canonicalizePath(self.filename)

    def fullSpan(self) -> Optional[tuple[int, int, int, int]]:
        if self.startCol and self.endLine and self.endCol:
            return (self.startLine, self.startCol, self.endLine, self.endCol)
        else:
            return None

    def code(self) -> Optional[str]:
        match self.fullSpan():
            case None:
                return None
            case (startLine, startCol, endLine, endCol):
                result = []
                for lineNo in range(startLine, startLine+1):
                    line = getline(self.filename, lineNo)
                    c1 = startCol if lineNo == startLine else 0
                    c2 = endCol if lineNo == endLine else len(line)
                    result.append(line[c1:c2])
                return '\n'.join(result)

    @staticmethod
    def fromFrameInfo(fi: inspect.FrameInfo) -> Loc:
        default = Loc(fi.filename, fi.lineno, None, None, None)
        if fi.positions is None:
            return default
        p: dis.Positions = fi.positions
        startLine = p.lineno
        endLine = p.end_lineno
        startCol = p.col_offset
        endCol = p.end_col_offset
        if startLine is None or endLine is None or startCol is None or endCol is None:
            return default
        else:
            return Loc(fi.filename, startLine, startCol, endLine, endCol)

HIGHLIGHTING_ENV_VAR = 'WYPP_HIGHLIGHTING'
type HighlightMode = Literal['color', 'text', 'off']

def getHighlightMode(mode: HighlightMode | Literal['fromEnv']) -> HighlightMode:
    if mode == 'fromEnv':
        fromEnv = utils.getEnv(HIGHLIGHTING_ENV_VAR, lambda x: x, None)
        match fromEnv:
            case 'color': return 'color'
            case 'text': return 'text'
            case 'off': return 'off'
            case None: return 'color'
            case _:
                myLogging.warn(f'Invalid highlighting mode in environment variable {HIGHLIGHTING_ENV_VAR}: {fromEnv} (supported: color, text, off)')
                return 'off'
    else:
        return mode

def highlight(s: str, mode: HighlightMode) -> str:
    match mode:
        case 'color': return ansi.red(s)
        case 'off': return s
        case 'text': return f'<<{s}>>'

@dataclass
class SourceLine:
    line: EncodedBytes               # without trailing \n
    span: Optional[tuple[int, int]]  # (inclusive, exclusive)

    def highlight(self, mode: HighlightMode | Literal['fromEnv'] = 'fromEnv') -> str:
        mode = getHighlightMode(mode)
        if self.span:
            l = self.line
            return l[:self.span[0]] + highlight(l[self.span[0]:self.span[1]], mode) + l[self.span[1]:]
        else:
            return self.line.decoded()

def highlightedLines(loc: Loc) -> list[SourceLine]:
    match loc.fullSpan():
        case None:
            line = getline(loc.filename, loc.startLine)
            return [SourceLine(line, None)]
        case (startLine, startCol, endLine, endCol):
            result = []
            for lineNo in range(startLine, startLine+1):
                line = getline(loc.filename, lineNo)
                leadingSpaces = line.countLeadingSpaces()
                c1 = startCol if lineNo == startLine else leadingSpaces
                c2 = endCol if lineNo == endLine else len(line)
                result.append(SourceLine(line, (c1, c2)))
            return result

@dataclass
class ClassMember:
    kind: Literal['method', 'recordConstructor']
    className: str

type CallableKind = Literal['function'] | ClassMember

@dataclass
class CallableName:
    name: str
    kind: CallableKind
    @staticmethod
    def mk(c: CallableInfo) -> CallableName:
        return CallableName(c.name, c.kind)

class CallableInfo(abc.ABC):
    """
    Class giving access to various properties of a function, method or constructor.
    """
    def __init__(self, kind: CallableKind):
        self.kind: CallableKind = kind
    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass
    @abc.abstractmethod
    def getResultTypeLocation(self) -> Optional[Loc]:
        pass
    @abc.abstractmethod
    def getParamSourceLocation(self, paramName: str) -> Optional[Loc]:
        pass

class StdCallableInfo(CallableInfo):
    """
    Class giving access to various properties of a function
    (arguments, result type etc.)
    """
    def __init__(self, f: Callable, kind: CallableKind):
        super().__init__(kind)
        self.file = f.__code__.co_filename
        self.__lineno = f.__code__.co_firstlineno
        self.__name = f.__name__
        self.__ast = parsecache.getAST(self.file)

    def __repr__(self):
        return f'StdCallableInfo({self.name}, {self.kind})'

    @property
    def name(self):
        return self.__name

    def _findDef(self) -> Optional[ast.FunctionDef | ast.AsyncFunctionDef]:
        m = FunMatcher(self.__name, self.__lineno)
        match self.kind:
            case 'function':
                return self.__ast.getFunDef(m)
            case ClassMember('method', clsName):
                return self.__ast.getMethodDef(clsName, m)
            case k:
                raise ValueError(f'Unexpected CallableKind {k} in StdCallableInfo')

    def getResultTypeLocation(self) -> Optional[Loc]:
        """
        Returns the location of the result type
        """
        node = self._findDef()
        if not node:
            return None
        r = node.returns
        if r:
            return Loc(self.file,
                       r.lineno,
                       r.col_offset,
                       r.end_lineno,
                       r.end_col_offset)
        else:
            # There is no return type annotation
            return None

    def getParamSourceLocation(self, paramName: str) -> Optional[Loc]:
        """
        Returns the location of the parameter with the given name.
        """
        node = self._findDef()
        if not node:
            return None
        res = None
        for arg in node.args.args + node.args.kwonlyargs:
            if arg.arg == paramName:
                res = arg
                break
        if res is None:
            # Look in vararg and kwarg
            if node.args.vararg and node.args.vararg.arg == paramName:
                res = node.args.vararg
            if node.args.kwarg and node.args.kwarg.arg == paramName:
                res = node.args.kwarg
        if res is None:
            return None
        else:
            return Loc(self.file,
                       res.lineno,
                       res.col_offset,
                       res.end_lineno,
                       res.end_col_offset)


def classFilename(cls) -> str | None:
    """Best-effort path to the file that defined `cls`."""
    try:
        fn = inspect.getsourcefile(cls) or inspect.getfile(cls)
        if fn:
            return fn
    except TypeError:
        pass
    # Fallback via the owning module (works for some frozen/zip cases)
    mod = sys.modules.get(cls.__module__)
    if mod is not None:
        return getattr(mod, "__file__", None) or getattr(getattr(mod, "__spec__", None), "origin", None)
    return None

class RecordConstructorInfo(CallableInfo):
    """
    Class giving access to various properties of a record constructor.
    """
    def __init__(self, cls: type):
        super().__init__(ClassMember('recordConstructor', cls.__name__))
        self.__cls = cls
    def __repr__(self):
        return f'RecordConstructorInfo({self.name})'
    @property
    def name(self):
        return self.__cls.__name__
    def getResultTypeLocation(self) -> Optional[Loc]:
        return None
    def getParamSourceLocation(self, paramName: str) -> Optional[Loc]:
        file = classFilename(self.__cls)
        if not file:
            return None
        ast = parsecache.getAST(file)
        node = ast.getRecordAttr(self.name, paramName)
        if node:
            return Loc(file, node.lineno, node.col_offset, node.end_lineno, node.end_col_offset)
        else:
            return None

def locationOfArgument(fi: inspect.FrameInfo, idxOrName: int | str) -> Optional[Loc]:
    """
    Given a stack frame with a function call f(arg1, arg2, ..., argN), returns
    the source code location of the i-th argument.
    """
    loc = Loc.fromFrameInfo(fi)
    match loc.fullSpan():
        case (startLine, startCol, _endLine, _endCol):
            codeOfCall = loc.code()
            if codeOfCall is None:
                return loc
            try:
                tree = ast.parse(codeOfCall)
            except SyntaxError:
                return loc
            match tree:
                case ast.Module([ast.Expr(ast.Call(_fun, args, kwArgs))]):
                    arg = None
                    if isinstance(idxOrName, int):
                        idx = idxOrName
                        if idx >= 0 and idx < len(args):
                            arg = args[idx]
                    else:
                        matching = [k.value for k in kwArgs if k.arg == idxOrName]
                        if matching:
                            arg = matching[0]
                    if arg is not None:
                        if arg.end_lineno is not None and arg.end_col_offset is not None:
                            callStartLine = startLine + arg.lineno - 1
                            callStartCol = startCol + arg.col_offset
                            callEndLine = startLine + arg.end_lineno - 1
                            if arg.lineno != arg.end_lineno:
                                callEndCol = arg.end_col_offset
                            else:
                                callEndCol = startCol + arg.end_col_offset
                            return Loc(loc.filename, callStartLine, callStartCol,
                                            callEndLine, callEndCol)
    return loc



import collections.abc
import types
from typing import *
import myTypeguard
#def renderTy(t: Any) -> str:
#    if isinstance(t, str):
#        return t
#    return str(t)
#    # return typeguard._utils.get_type_name(t, ['__wypp__'])

_NoneType = type(None)

def renderTy(tp: Any) -> str:
    # Does not work: return typeguard._utils.get_type_name(t, ['__wypp__'])
    # For example, Callable[[int, bool], str] is formated as "Callable[list, str]"

    if isinstance(tp, str):
        # forward reference
        return tp

    origin = get_origin(tp)
    args   = get_args(tp)

    # Simple / builtin / classes
    if origin is None:
        # e.g. int, str, custom classes, Any, typing.NoReturn, typing.Never
        if tp is Any: return "Any"
        if tp is _NoneType: return "None"
        if tp is types.EllipsisType: return "..."
        if isinstance(tp, list): return str(tp)
        if isinstance(tp, tuple): return str(tp)
        if isinstance(tp, dict): return str(tp)
        return myTypeguard.getTypeName(tp)

    # Union / Optional (PEP 604)
    if origin is Union or origin is types.UnionType:
        if len(args) == 2:
            if args[0] is _NoneType and args[1] is not _NoneType:
                return f"Optional[{renderTy(args[1])}]"
            elif args[1] is _NoneType and args[0] is not _NoneType:
                return f"Optional[{renderTy(args[0])}]"
        parts = [renderTy(a) for a in args]
        return "Union[" + ", ".join(parts) + "]"

    # Annotated[T, ...]
    if origin is Annotated:
        base, *meta = args
        if len(meta) >= 1 and isinstance(meta[-1], str):
            return meta[-1]
        metas = ", ".join(repr(m) for m in meta)
        return f"Annotated[{renderTy(base)}, {metas}]"

    # Literal[...]
    if origin is Literal:
        return "Literal[" + ", ".join(repr(a) for a in args) + "]"

    # Callable[[A, B], R] or Callable[..., R]
    if origin in (Callable, collections.abc.Callable):
        params, ret = args
        if params is Ellipsis:
            params_s = "..."
        else:
            params_s = ", ".join(renderTy(p) for p in params)
        return f"Callable[[{params_s}], {renderTy(ret)}]"

    # Tuple[T1, T2] or Tuple[T, ...]
    if origin is tuple:
        if len(args) == 2 and args[1] is Ellipsis:
            return f"tuple[{renderTy(args[0])}, ...]"
        elif len(args) == 0:
            return "tuple[()]"
        return "tuple[" + ", ".join(renderTy(a) for a in args) + "]"

    # ClassVar[T], Final[T]
    if origin is ClassVar:
        return f"ClassVar[{renderTy(args[0])}]"
    if origin is Final:
        return f"Final[{renderTy(args[0])}]"

    # Parametrized generics like list[T], dict[K, V], set[T], type[T], etc.
    name = getattr(origin, "__name__", None) or getattr(origin, "__qualname__", repr(origin))
    if args:
        return f"{name}[" + ", ".join(renderTy(a) for a in args) + "]"
    else:
        return name



from typing import *
import os
import sys
from contextlib import contextmanager

P = ParamSpec("P")
T = TypeVar("T")

# The name of this function is magical. All stack frames that appear
# nested within this function are removed from tracebacks.
def _call_with_frames_removed(
    f: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    return f(*args, **kwargs)

# The name of this function is magical. The next stack frame that appears
# nested within this function is removed from tracebacks.
def _call_with_next_frame_removed(
    f: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    return f(*args, **kwargs)

def getEnv(name, conv, default):
    s = os.getenv(name)
    if s is None:
        return default
    try:
        return conv(s)
    except:
        return default

def dropWhile(l: list, f: Callable[[Any], bool]) -> list:
    if not l:
        return l
    for i in range(len(l)):
        if not f(l[i]):
            break
    else:
        # All elements satisfied the condition, return empty list
        return []
    return l[i:]

def split(l: list, f: Callable[[Any], bool]) -> tuple[list, list]:
    """
    span, applied to a list l and a predicate f, returns a tuple where first element is the
    longest prefix (possibly empty) of l of elements that satisfy f and second element
    is the remainder of the list.
    """
    inFirst = True
    first = []
    second = []
    for x in l:
        if inFirst:
            if f(x):
                first.append(x)
            else:
                inFirst = False
                second.append(x)
        else:
            second.append(x)
    return (first, second)


def isUnderTest() -> bool:
    x = os.getenv('WYPP_UNDER_TEST')
    return x == 'True'

@contextmanager
def underTest(value: bool = True):
    """Context manager to temporarily set WYPP_UNDER_TEST environment variable."""
    oldValue = os.getenv('WYPP_UNDER_TEST')
    os.environ['WYPP_UNDER_TEST'] = str(value)
    try:
        yield
    finally:
        if oldValue is None:
            # Remove the environment variable if it didn't exist before
            os.environ.pop('WYPP_UNDER_TEST', None)
        else:
            # Restore the original value
            os.environ['WYPP_UNDER_TEST'] = oldValue


def die(ecode: str | int | None = 1):
    if isinstance(ecode, str):
        sys.stderr.write(ecode)
        sys.stderr.write('\n')
        ecode = 1
    elif ecode == None:
        ecode = 0
    if sys.flags.interactive:
        os._exit(ecode)
    else:
        sys.exit(ecode)

def readFile(path):
    try:
        with open(path, encoding='utf-8') as f:
            return f.read()
    except UnicodeDecodeError:
        with open(path) as f:
            return f.read()



import typing
from typing import Any
import dataclasses
import inspect
import errors
import typecheck
import records
import stacktrace
import renderTy
import location
import paths
import utils
import i18n

_DEBUG = False
def _debug(s):
    if _DEBUG:
        print('[DEBUG] ' + s)

record = records.record

intPositive = typing.Annotated[int, lambda i: i > 0, 'intPositive']
nat = typing.Annotated[int, lambda i: i >= 0, 'nat']
intNonNegative = typing.Annotated[int, lambda i: i >= 0, 'intNonNegative']
intNonPositive = typing.Annotated[int, lambda i: i <= 0, 'intNonPositive']
intNegative = typing.Annotated[int, lambda i: i < 0, 'intNegative']

floatPositive = typing.Annotated[float, lambda x: x > 0, 'floatPositive']
floatNonNegative = typing.Annotated[float, lambda x: x >= 0, 'floatNonNegative']
floatNegative = typing.Annotated[float, lambda x: x < 0, 'floatNegative']
floatNonPositive = typing.Annotated[float, lambda x: x <= 0, 'floatNonPositive']

class Lock(typing.Protocol):
    def acquire(self, blocking: bool = True, timeout:int = -1) -> Any:
       pass
    def release(self) -> Any:
        pass
    def locked(self) -> Any:
        pass

LockFactory = typing.Annotated[typing.Callable[[], Lock], 'LockFactory']

T = typing.TypeVar('T')
U = typing.TypeVar('U')
V = typing.TypeVar('V')

# Dirty hack ahead: we patch some methods of internal class of the typing module.

def _literalInstanceOf(self, value):
    for p in self.__args__:
        # typing.Literal checks for type(value) == type(arg),
        # however this does not work well with untypy.
        if isinstance(value, type(p)) and p == value:
            return True
    return False

# This patch is needed to be able to use a literal L to check whether a value x
# is an instance of this literal: isinstance(x, L)
# pyright does not know about typing._LiteralGenericAlias, we do not typecheck the following line.
setattr(typing._LiteralGenericAlias, '__instancecheck__', _literalInstanceOf) # type: ignore

def _invalidCall(self, *args, **kwds):
    if hasattr(self, '__name__'):
        name = self.__name__
    else:
        name = str(self)
        typingPrefix = 'typing.'
        if name.startswith(typingPrefix):
            name = name[len(typingPrefix):]
    def formatArg(x):
        if name == 'Literal':
            return repr(x)
        else:
            return renderTy.renderTy(x)
    caller = stacktrace.callerOutsideWypp()
    loc = None if caller is None else location.Loc.fromFrameInfo(caller)
    argStr = ', '.join([formatArg(x) for x in args])
    tyStr = f'{name}({argStr})'
    raise utils._call_with_frames_removed(errors.WyppTypeError.invalidType, tyStr, loc)

# This patch is needed to provide better error messages if a student passes type arguments
# with paranthesis instead of square brackets
setattr(typing._SpecialForm, '__call__', _invalidCall)

def typechecked(func=None):
    def wrap(func):
        if hasattr(func, '__qualname__'):
            q = getattr(func, "__qualname__").split('.')
            if len(q) >= 2:
                className = q[-2]
                cfg = {'kind': 'method', 'className': className}
            else:
                cfg = {'kind': 'function'}
        else:
            cfg = {'kind': 'function'}
        return typecheck.wrapTypecheck(cfg)(func)
    if func is None:
        # We're called with parens.
        return wrap
    else:
        # We're called as @dataclass without parens.
        return wrap(func)

# Tests

_die = False

def setDieOnCheckFailures(b):
    global _die
    _die = b

def _dieOnCheckFailures():
    return _die

_testCount = {'total': 0, 'failing': 0}
_checksEnabled = True
_typeCheckingEnabled = False

def initModule(enableChecks=True, enableTypeChecking=True, quiet=False):
    global _checksEnabled, _typeCheckingEnabled
    _checksEnabled = enableChecks
    _typeCheckingEnabled = enableTypeChecking
    records.init(enableTypeChecking)
    resetTestCount()

def resetTestCount():
    global _testCount
    _testCount = {'total': 0, 'failing': 0}

def printTestResults(prefix='', loadingFailed=False):
    total = _testCount['total']
    failing = _testCount['failing']
    bad = '🙁'
    good = '😀'
    tests = f'{prefix}' + i18n.numTests(total)
    if total == 0:
        pass
    elif failing == 0:
        if loadingFailed:
            print(f'{tests}, {i18n.tr("Stop of execution")} {bad}')
        elif total == 1:
            print(f'{i18n.tr("1 successful test")} {good}')
        else:
            print(f'{tests}, {i18n.tr("all successful")} {good}')
    else:
        if loadingFailed:
            print(f'{tests}, {i18n.numFailing(failing)} and stop of execution {bad}')
        else:
            print(f'{tests}, {i18n.numFailing(failing)} {bad}')
    return {'total': total, 'failing': failing}

def checkEq(actual, expected):
    return checkGeneric(actual, expected, structuralObjEq=False)

def incTestCount(testOk: bool):
    global _testCount
    _testCount = {
        'total': _testCount['total'] + 1,
        'failing': _testCount['failing'] + (0 if testOk else 1)
    }

def check(actual, expected):
    """
    Überprüft, ob ein Funktionsaufruf das gewünschte Ergebnis liefert.

    Beispiel:

    `check(f(10, 'blah'), 17)` überprüft, ob der Aufruf `f(10, 'blah')` wie erwartet das Ergebnis `17` liefert.
    """
    checkGeneric(actual, expected)

def checkGeneric(actual, expected, *, structuralObjEq=True, floatEqWithDelta=True):
    if not _checksEnabled:
        return
    flags = {'structuralObjEq': structuralObjEq, 'floatEqWithDelta': floatEqWithDelta}
    matches = deepEq(actual, expected, **flags)
    incTestCount(matches)
    if not matches:
        stack = inspect.stack()
        frame = stack[2] if len(stack) > 2 else None
        if frame:
            filename = paths.canonicalizePath(frame.filename)
            caller = i18n.tr('File {filename}, line {lineno}: ',
                             filename=filename, lineno=frame.lineno)
        else:
            caller = ""
        def fmt(x):
            if type(x) == str:
                return repr(x)
            else:
                return str(x)
        msg = caller + i18n.checkExpected(fmt(expected), fmt(actual))
        if _dieOnCheckFailures():
            raise Exception(msg)
        else:
            print(i18n.tr('ERROR in ') + msg)

def checkFail(msg: str):
    if not _checksEnabled:
        return
    incTestCount(False)
    msg = str(msg)
    if _dieOnCheckFailures():
        raise Exception(msg)
    else:
        print(i18n.tr('ERROR: ') + msg)

def uncoveredCase():
    stack = inspect.stack()
    if len(stack) > 1:
        caller = stack[1]
        callerStr = i18n.tr('File {filename}, line {lineno}: ',
                            filename=caller.filename, lineno=caller.lineno)
        raise Exception(callerStr + i18n.tr('uncovered case'))
    else:
        raise Exception(i18n.tr('Uncovered case'))

#
# Deep equality
#
def _isNumber(x):
    t = type(x)
    return (t is int or t is float)

def _seqEq(seq1, seq2, flags):
    if len(seq1) != len(seq2):
        return False
    for i, x in enumerate(seq1):
        y = seq2[i]
        if not deepEq(x, y, **flags):
            return False
    return True

def _dictEq(d1, d2, flags):
    ks1 = set(d1.keys())
    ks2 = set(d2.keys())
    if ks1 != ks2: # keys should be exactly equal
        return False
    for k in ks1:
        if not deepEq(d1[k], d2[k], **flags):
            return False
    return True

def _objToDict(o):
    d = {}
    attrs = dir(o)
    useAll = False
    if hasattr(o, records.EQ_ATTRS_ATTR):
        useAll = True
        attrs = getattr(o, records.EQ_ATTRS_ATTR)
    for n in attrs:
        if not useAll and n and n[0] == '_':
            continue
        x = getattr(o, n)
        d[n] = x
    #print(f'_objToDict: {d}')
    return d

def _objEq(o1, o2, flags):
    return _dictEq(_objToDict(o1), _objToDict(o2), flags)

_EPSILON = 0.00001

def _useFloatEqWithDelta(flags):
    return flags.get('floatEqWithDelta', False)

def _useStructuralObjEq(flags):
    return flags.get('structuralObjEq', False)

# Supported flags:
# - structuralObjEq: causes objects to be compared as dictionaries. The keys are by default
#   taken from dir(obj), if obj as the attribute __eqAttrs__, then the names listed there
#   are taken. Defaults to False.
# - floatEqWithDelta: compares floats by checking whether the difference is smaller than a
#   small delta. Setting this to True loses transitivity of eq.
def deepEq(v1, v2, **flags):
    """
    Computes deep equality of v1 and v2. With structuralObjEq=False, objects are compared
    by __eq__. Otherwise, objects are compared attribute-wise, only those attributes
    returned by dir that do not start with an underscore are compared.
    """
    #print(f'deepEq: v1={v1}, v2={v2}, flags={flags}')
    if v1 == v2:
        return True
    if _isNumber(v1) and _isNumber(v2) and _useFloatEqWithDelta(flags):
        diff = v1 - v2
        if abs(diff) < _EPSILON:
            return True
        else:
            return False
    if isinstance(v1, list):
        return isinstance(v2, list) and _seqEq(v1, v2, flags)
    if isinstance(v1, tuple):
        return isinstance(v2, tuple) and _seqEq(v1, v2, flags)
    if isinstance(v1, dict):
        return isinstance(v2, dict) and _dictEq(v1, v2, flags)
    if callable(v1):
        return False
    if type(v1) == str:
        return False
    if hasattr(v1, '__class__'):
        if _useStructuralObjEq(flags):
            res = _objEq(v1, v2, flags)
            return res
        else:
            return False # v1 == v2 already checked
    return False

def todo(msg=None):
    if msg is None:
        msg = 'TODO'
    raise errors.TodoError(msg)

def impossible(msg=None):
    if msg is None:
        msg = i18n.tr('The impossible happened!')
        'Das Unmögliche ist passiert!'
    raise errors.ImpossibleError(msg)

# Additional functions and aliases

import math as moduleMath

math = moduleMath

wrapTypecheck = typecheck.wrapTypecheck



import sys
import os
import importlib
import runpy
from dataclasses import dataclass

# local imports
from constants import *
import stacktrace
import instrument
from myLogging import *
from exceptionHandler import handleCurrentException
import utils

class Lib:
    def __init__(self, mod, properlyImported):
        self.properlyImported = properlyImported
        if not properlyImported:
            self.initModule = mod['initModule']
            self.resetTestCount = mod['resetTestCount']
            self.printTestResults = mod['printTestResults']
            self.dict = mod
        else:
            self.initModule = mod.initModule
            self.resetTestCount = mod.resetTestCount
            self.printTestResults = mod.printTestResults
            d = {}
            self.dict = d
            for name in dir(mod):
                if name and name[0] != '_':
                  d[name] = getattr(mod, name)


@dataclass
class RunSetup:
    def __init__(self, pathDir: str, args: list[str]):
        self.pathDir = pathDir
        self.args = args
        self.sysPathInserted = False
        self.oldArgs = sys.argv
    def __enter__(self):
        if self.pathDir not in sys.path:
            sys.path.insert(0, self.pathDir)
            self.sysPathInserted = True
        sys.argv = self.args
        self.originalProfile = sys.getprofile()
        stacktrace.installProfileHook()
    def __exit__(self, exc_type, value, traceback):
        sys.setprofile(self.originalProfile)
        if self.sysPathInserted:
            sys.path.remove(self.pathDir)
            self.sysPathInserted = False
        sys.argv = self.oldArgs

def prepareLib(onlyCheckRunnable, enableTypeChecking):
    libDefs = None
    mod = 'wypp'
    verbose('Attempting to import ' + mod)
    wypp = importlib.import_module(mod)
    libDefs = Lib(wypp, True)
    verbose(f'Successfully imported module {mod} from file {wypp.__file__}')
    libDefs.initModule(enableChecks=not onlyCheckRunnable,
                       enableTypeChecking=enableTypeChecking,
                       quiet=onlyCheckRunnable)
    return libDefs

def runCode(fileToRun, globals, doTypecheck=True, extraDirs=None) -> dict:
    if not extraDirs:
        extraDirs = []
    modName = os.path.basename(os.path.splitext(fileToRun)[0])
    with instrument.setupFinder(os.path.dirname(fileToRun), modName, extraDirs, doTypecheck):
        sys.dont_write_bytecode = True
        res = runpy.run_module(modName, init_globals=globals, run_name='__wypp__', alter_sys=False)
        return res

def runStudentCode(fileToRun, globals, onlyCheckRunnable, doTypecheck=True, extraDirs=None) -> dict:
    doRun = lambda: runCode(fileToRun, globals, doTypecheck=doTypecheck, extraDirs=extraDirs)
    if onlyCheckRunnable:
        try:
            doRun()
        except:
            printStderr('Loading file %s crashed' % fileToRun)
            handleCurrentException()
        else:
            utils.die(0)
    return doRun()

# globals already contain libDefs
def runTestsInFile(testFile, globals, libDefs, doTypecheck=True, extraDirs=[]):
    printStderr()
    printStderr(f"Running tutor's tests in {testFile}")
    libDefs.resetTestCount()
    try:
        runCode(testFile, globals, doTypecheck=doTypecheck, extraDirs=extraDirs)
    except:
        handleCurrentException()
    return libDefs.dict['printTestResults']('Tutor:  ')

# globals already contain libDefs
def performChecks(check, testFile, globals, libDefs, doTypecheck=True, extraDirs=None, loadingFailed=False):
    prefix = ''
    if check and testFile:
        prefix = 'Student: '
    testResultsStudent = libDefs.printTestResults(prefix, loadingFailed)
    if check:
        testResultsInstr = {'total': 0, 'failing': 0}
        if testFile:
            testResultsInstr = runTestsInFile(testFile, globals, libDefs, doTypecheck=doTypecheck,
                                              extraDirs=extraDirs)
        failingSum = testResultsStudent['failing'] + testResultsInstr['failing']
        utils.die(0 if failingSum < 1 else 1)



RESET         = "\u001b[0;0m"
BOLD          = "\u001b[1m"
REVERSE       = "\u001b[2m"

BLACK         = "\u001b[0;30m"
BLUE          = "\u001b[0;34m"
GREEN         = "\u001b[0;32m"
CYAN          = "\u001b[0;36m"
RED           = "\u001b[0;31m"
PURPLE        = "\u001b[0;35m"
BROWN         = "\u001b[0;33m"
GRAY          = "\u001b[0;37m"
DARK_GRAY     = "\u001b[1;30m"
LIGHT_BLUE    = "\u001b[1;34m"
LIGHT_GREEN   = "\u001b[1;32m"
LIGHT_CYAN    = "\u001b[1;36m"
LIGHT_RED     = "\u001b[1;31m"
LIGHT_PURPLE  = "\u001b[1;35m"
YELLOW        = "\u001b[1;33m"
WHITE         = "\u001b[1;37m"

def color(s, color):
    return color + s + RESET

def green(s):
    return color(s, GREEN)

def red(s):
    return color(s, RED + BOLD)

def blue(s):
    return color(s, BLUE + BOLD)



from typing import *
import os
import ast
import importlib
import importlib.abc
from importlib.machinery import ModuleSpec, SourceFileLoader
import importlib.machinery
from importlib.util import decode_source, spec_from_file_location
from collections.abc import Buffer
import types
from os import PathLike
import utils
from myLogging import *
from contextlib import contextmanager
import errors
import location

def parseExp(s: str) -> ast.expr:
    match ast.parse(s):
        case ast.Module([ast.Expr(e)], type_ignores):
            return e
        case m:
            raise ValueError(f'String {repr(s)} does not parse as an expression: {m}')

class Configs:
    funConfig: ast.expr = parseExp("{'kind': 'function', 'globals': globals(), 'locals': locals()}")
    @staticmethod
    def methodConfig(clsName: str) -> ast.expr:
        return parseExp("{'kind': 'method', 'className': " + repr(clsName) + ", 'globals': globals(), 'locals': locals()}")
    immutableRecordConfig: ast.expr = parseExp('record(mutable=False, globals=globals(), locals=locals())')
    mutableRecordConfig: ast.expr = parseExp('record(mutable=True, globals=globals(), locals=locals())')

def transferLocs(old: ast.stmt | ast.expr, new: ast.stmt | ast.expr) -> Any:
    new.lineno = old.lineno
    new.col_offset = old.col_offset
    new.end_lineno = old.end_lineno
    new.end_col_offset = old.end_col_offset
    return new

def transformDecorator(e: ast.expr, path: str) -> ast.expr:
    loc = location.Loc(path, e.lineno, e.col_offset, e.end_lineno, e.col_offset)
    match e:
        case ast.Name('record'):
            return transferLocs(e, Configs.immutableRecordConfig)
        case ast.Call(ast.Name('record'), [], kwArgs):
            match kwArgs:
                case [ast.keyword('mutable', ast.Constant(True))]:
                    return transferLocs(e, Configs.mutableRecordConfig)
                case [ast.keyword('mutable', ast.Constant(False))]:
                    return transferLocs(e, Configs.immutableRecordConfig)
                case _:
                    raise errors.WyppTypeError.invalidRecordAnnotation(loc)
        case ast.Call(ast.Name('record'), _, _):
            raise ValueError(f'Invalid record config')
        case _:
            return e

def transformStmt(stmt: ast.stmt, outerClassName: Optional[str], path: str) -> ast.stmt:
    cfg = Configs.methodConfig(outerClassName) if outerClassName else Configs.funConfig
    wrapExp = ast.Call(ast.Name(id='wrapTypecheck', ctx=ast.Load()), [cfg], [])
    match stmt:
        case ast.FunctionDef(name, args, body, decorators, returns, tyComment, tyParams):
            newBody = [transformStmt(s, outerClassName=outerClassName, path=path) for s in body]
            x = ast.FunctionDef(name, args, newBody, decorators + [wrapExp], returns, tyComment, tyParams)
            return transferLocs(stmt, x)
        case ast.AsyncFunctionDef(name, args, body, decorators, returns, tyComment, tyParams):
            newBody = [transformStmt(s, outerClassName=outerClassName, path=path) for s in body]
            x = ast.AsyncFunctionDef(name, args, newBody, decorators + [wrapExp], returns, tyComment, tyParams)
            return transferLocs(stmt, x)
        case ast.ClassDef(className, bases, keywords, body, decoratorList, type_params):
            newBody = [transformStmt(s, outerClassName=className, path=path) for s in body]
            newDecoratorList = [transformDecorator(e, path=path) for e in decoratorList]
            x = ast.ClassDef(className, bases, keywords, newBody, newDecoratorList, type_params)
            return transferLocs(stmt, x)
        case _:
            return stmt

def isImport(t: ast.stmt) -> bool:
    match t:
        case ast.Import(): return True
        case ast.ImportFrom(): return True
        case _: return False

importWrapTypecheck = ast.parse("from wypp import wrapTypecheck", mode="exec").body[0]

def transformModule(m: ast.Module | ast.Expression | ast.Interactive, path: str) -> ast.Module | ast.Expression | ast.Interactive:
    match m:
        case ast.Module(body, type_ignores):
            newStmts = [transformStmt(stmt, outerClassName=None, path=path) for stmt in body]
            (imports, noImports) = utils.split(newStmts, isImport)
            # avoid inserting before from __future__
            newStmts = imports + [importWrapTypecheck] + noImports
            return ast.Module(newStmts, type_ignores)
        case _:
            return m

class InstrumentingLoader(SourceFileLoader):
    @staticmethod
    def source_to_code(
        data: Buffer | str | ast.Module | ast.Expression | ast.Interactive,
        path: Buffer | str | PathLike[str] = "<string>",
    ) -> types.CodeType:
        if isinstance(data, (ast.Module, ast.Expression, ast.Interactive)):
            tree = data
        else:
            if isinstance(data, str):
                source = data
            else:
                source = decode_source(data)
            tree = utils._call_with_frames_removed(ast.parse, source, path, "exec")
        if isinstance(path, PathLike):
            pathStr = str(path)
        elif isinstance(path, str):
            pathStr = path
        else:
            pathStr = "<input>"
        tree = transformModule(tree, pathStr)
        ast.fix_missing_locations(tree)

        debug(
            f"Source code of {path!r} after instrumentation:\n" +
            "----------------------------------------------\n" +
            ast.unparse(tree) + "\n"
            "----------------------------------------------")

        code = utils._call_with_frames_removed(compile, tree, path, "exec", 0, dont_inherit=True)
        return code

class InstrumentingFinder(importlib.abc.MetaPathFinder):
    def __init__(self, finder, modDir: str, modName: str, extraDirs: list[str]):
        self._origFinder = finder
        self.mainModName = modName
        self.modDir = os.path.realpath(modDir) + '/'
        self.extraDirs = [os.path.realpath(d) for d in extraDirs]

    def find_spec(
            self,
            fullname: str,
            path: Sequence[str] | None,
            target: types.ModuleType | None = None,
        ) -> ModuleSpec | None:

        # 1) The fullname is the name of the main module. This might be a dotted name such as x.y.z.py
        #    so we have special logic here
        fp = os.path.join(self.modDir, f"{fullname}.py")
        if self.mainModName == fullname and os.path.isfile(fp):
            loader = InstrumentingLoader(fullname, fp)
            return spec_from_file_location(fullname, fp, loader=loader)
        # 2) The fullname is a prefix of the main module. We want to load main modules with
        #    dotted names such as x.y.z.py, hence we synthesize a namespace pkg
        #    e.g. if 'x.y.z.py' exists and we're asked for 'x', return a package spec.
        elif self.mainModName.startswith(fullname + '.'):
            spec = importlib.machinery.ModuleSpec(fullname, loader=None, is_package=True)
            # Namespace package marker (PEP 451)
            spec.submodule_search_locations = []
            return spec
        # 3) Fallback: use the original PathFinder
        spec = self._origFinder.find_spec(fullname, path, target)
        debug(f'spec for {fullname}: {spec}')
        if spec is None:
            return spec

        origin = os.path.realpath(spec.origin)
        dirs = [self.modDir] + self.extraDirs
        isLocalModule = False
        for d in dirs:
            if origin.startswith(d):
                isLocalModule = True
                break
        # print(f'Module {fullname} is locale: {isLocalModule} ({origin})')
        if spec and spec.loader and isLocalModule:
            spec.loader = InstrumentingLoader(spec.loader.name, spec.loader.path)
        return spec

@contextmanager
def setupFinder(modDir: str, modName: str, extraDirs: list[str], typechecking: bool):
    if not typechecking:
        yield
    else:
        # Find the PathFinder
        for finder in sys.meta_path:
            if (
                isinstance(finder, type)
                and finder.__name__ == "PathFinder"
                and hasattr(finder, "find_spec")
            ):
                break
        else:
            raise RuntimeError("Cannot find a PathFinder in sys.meta_path")

        # Create and install our custom finder
        instrumenting_finder = InstrumentingFinder(finder, modDir, modName, extraDirs)
        sys.meta_path.insert(0, instrumenting_finder)

        try:
            yield
        finally:
            # Remove our custom finder when exiting the context
            if instrumenting_finder in sys.meta_path:
                sys.meta_path.remove(instrumenting_finder)



from dataclasses import dataclass
import location
from typing import *
from contextlib import contextmanager
import lang
import utils

type Lang = Literal['en', 'de']

allLanguages: list[Lang] = ['en', 'de']

# If not set explicitly, the current locale is asked for the language
_lang: Optional[Lang] = None

@contextmanager
def explicitLang(newLang: Lang):
    """Context manager to temporarily set the language."""
    global _lang
    oldLang = _lang
    _lang = newLang
    try:
        yield
    finally:
        _lang = oldLang

def getLang() -> Lang:
    if _lang:
        return _lang
    else:
        return lang.pickLanguage(allLanguages, 'en')

def setLang(lang: Lang):
    global _lang
    _lang = lang

def tr(key: str, **kws) -> str:
    match getLang():
        case 'en':
            tmpl = key
        case 'de':
            if key not in DE and utils.isUnderTest():
                raise ValueError(f'Untranslated string: {key}')
            tmpl = DE.get(key, key)
        case x:
            if utils.underTest():
                raise ValueError(f'Unknown language: {x}')
    return tmpl.format(**kws)

DE = {
    'Expecting no return value when calling function `{fun}`.':
        'Kein Rückgabewert erwartet bei Aufruf der Funktion `{fun}`.',
    'Expecting no return value when calling method `{method}` of class `{cls}`.':
        'Kein Rückgabewert erwartet bei Aufruf der Methode `{method}` der Klasse `{cls}`.',
    'Expecting no return value when calling constructor of record `{cls}`.':
        'Kein Rückgabewert erwartet bei Aufruf des Konstruktors des Records `{cls}`.',

    'Expecting return value of type `{ty}` when calling function `{fun}`.':
        'Rückgabewert vom Typ `{ty}` erwartet bei Aufruf der Funktion `{fun}`.',
    'Expecting return value of type `{ty}` when calling method `{method}` of class `{cls}`.':
        'Rückgabewert vom Typ `{ty}` erwartet bei Aufruf der Methode `{method}` der Klasse `{cls}`.',
    'Expecting return value of type `{ty}` when calling constructor of record `{cls}`.':
        'Rückgabewert vom Typ `{ty}` erwartet bei Aufruf des Konstruktors des Records `{cls}`.',

    'But the call returns a value of type `{ty}`.':
        'Aber der Aufruf gibt einen Wert vom Typ `{ty}` zurück.',

    'But no value returned.':
        'Aber kein Rückgabewert vorhanden.',
    'no result returned': 'kein Rückgabewert vorhanden',
    'Call in line {line} causes the function to return no value:':
        'Aufruf in Zeile {line} führt dazu, dass die Funktion keinen Wert zurückgibt:',

    'The call of function `{fun}` expects value of type `{ty}` as {arg}.':
        'Der Aufruf der Funktion `{fun}` erwartet einen Wert vom Typ `{ty}` als {arg}.',
    'The call of method `{method}` of class `{cls}` expects value of type `{ty}` as {arg}.':
        'Der Aufruf der Methode `{method}` der Klasse `{cls}` erwartet einen Wert vom Typ `{ty}` als {arg}.',
    'The call of the constructor of record `{cls}` expects value of type `{ty}` as {arg}.':
        'Der Aufruf des Konstruktors des Records `{cls}` erwartet einen Wert vom Typ `{ty}` als {arg}.',

    'But the value given has type `{ty}`.': 'Aber der übergebene Wert hat den Typ `{ty}`.',
    'But the default value has type `{ty}`.': 'Aber der Default-Wert hat den Typ `{ty}`.',

    'File': 'Datei',
    'Line': 'Zeile',
    'Result type declared in line': 'Rückgabetyp deklariert in Zeile',
    'Type declared in line': 'Typ deklariert in Zeile',
    'Parameter declared in line': 'Parameter deklariert in Zeile',
    'Problematic return in line': 'Fehlerhaftes return in Zeile',
    'Problematic call in line': 'Fehlerhafter Aufruf in Zeile',
    'Call in line {line} causes the problematic return:':
        'Aufruf in Zeile {line} verursacht das fehlerhafte return:',

    'Parameter `{param}` of function `{fun}` requires a type annotation.':
        'Parameter `{param}` der Funktion `{fun}` benötigt eine Typannotation.',
    'Parameter `{param}` of method `{method}` from class `{cls}` requires a type annotation.':
        'Parameter `{param}` der Methode `{method}` der Klasse `{cls}` benötigt eine Typannotation.',
    'Parameter `{param}` of constructor of record `{cls}` requires a type annotation.':
        'Parameter `{param}` des Konstruktors des Records `{cls}` benötigt eine Typannotation.',

    'Attribute `{name}` of record `{record}` requires a type annotation.':
        'Attribut `{name}` des Records `{record}` benötigt eine Typannotation.',

    'invalid type `{ty}`':
        'ungültiger Typ `{ty}`',
    'Cannot set attribute to value of type `{ty}`.':
        'Das Attribut kann nicht auf einen Wert vom Typ `{ty}` gesetzt werden.',
    'Problematic assignment in line': 'Fehlerhafte Zuweisung in Zeile',
    'Attribute `{attrName}` of record `{recordName}` declared with type `{ty}`.':
        'Attribut `{attrName}` des Records `{recordName}` deklariert als Typ `{ty}`.',

    'argument count mismatch': 'Anzahl der Argumente stimmt nicht',
    'Call in line': 'Aufruf in Zeile',
    'Function `{fun}` takes ': 'Funktion `{fun}` benötigt ',
    'Function `{fun}` takes at least ': 'Funktion `{fun}` benötigt mindestens ',
    'Function `{fun}` takes at most ': 'Funktion `{fun}` akzeptiert höchstens ',
    'Method `{method}` of class `{cls}` takes ': 'Methode `{method}` der Klasse `{cls}` benötigt ',
    'Method `{method}` of class `{cls}` takes at least ': 'Methode `{method}` der Klasse `{cls}` benötigt mindestens ',
    'Method `{method}` of class `{cls}` takes at most ': 'Methode `{method}` der Klasse `{cls}` akzeptiert höchstens ',
    'Constructor of record `{cls}` takes ': 'Konstruktor des Records `{cls}` benötigt ',
    'Constructor of record `{cls}` takes at least ': 'Konstruktor des Records `{cls}` benötigt mindestens ',
    'Constructor of record `{cls}` takes at most ': 'Konstruktor des Records `{cls}` akzeptiert höchstens ',
    'Given: ': 'Gegeben: ',

    'Default value for parameter `{paramName}` of function `{fun}` must have type `{ty}`.':
        'Default-Wert des Parameters `{paramName}` der Funktion `{fun}` muss vom Typ `{ty}` sein.',
    'Default value for parameter `{paramName}` of method `{method}` in class `{cls}` must have type `{ty}`.':
        'Default-Wert des Parameters `{paramName}` der Methode `{method}` der Klasse `{cls}` muss vom Typ `{ty}` sein.',
    'Default value for attribute `{paramName}` of record `{cls}` must have type `{ty}`.':
        'Default-Wert des Attributs `{paramName}` des Records `{cls}` muss vom Typ `{ty}` sein.',

    'Unknown attribute {attrName} for record {clsName}':
        'Attribut {attrName} ist nicht bekannt für den Record {clsName}',
    'Did you mean `{ty}`?': 'Wolltest du `{ty}` schreiben?',

    'unknown keyword argument': 'unbekanntes Schlüsselwort-Argument',
    'Function `{fun}` does not accept keyword argument `{name}`.':
        'Funktion `{fun}` akzeptiert kein Schlüsselwort-Argument `{name}`.',
    'Method `{method}` from class `{cls}` does not accept keyword argument `{name}`.':
        'Methode `{method}` der Klasse `{cls}` akzeptiert kein Schlüsselwort-Argument `{name}`.',
    'Constructor of record `{cls}` does not accept keyword argument `{name}`.':
        'Konstruktor des Records `{cls}` akzeptiert kein Schlüsselwort-Argument `{name}`.',

    'invalid record definition': 'ungültige Record-Definition',

    'Expected {expected}, but the result is {actual}':
        'Erwartet wird {expected}, aber das Ergebnis ist {actual}',
    'ERROR: ': 'FEHLER: ',
    'ERROR in ': 'FEHLER in ',
    'File {filename}, line {lineno}: ': 'Datei {filename}, Zeile {lineno}: ',
    'Uncovered case': 'Ein Fall ist nicht abgedeckt',
    'uncovered case': 'ein Fall ist nicht abgedeckt',
    'The impossible happened!': 'Das Unmögliche ist passiert!',
    'Stop of execution': 'Abbruch der Ausführung',
    '1 successful test': '1 erfolgreicher Test',
    'all successful': 'alle erfolgreich',
    'and stop of execution': 'und Abbruch der Ausführung',

    '=== WELCOME to ': '=== WILLKOMMEN bei '
}

def expectingNoReturn(cn: location.CallableName) -> str:
    match cn.kind:
        case 'function':
            return tr('Expecting no return value when calling function `{fun}`.',
                    fun=cn.name)
        case location.ClassMember('method', cls):
            return tr('Expecting no return value when calling method `{method}` of class `{cls}`.',
                    method=cn.name, cls=cls)
        case location.ClassMember('recordConstructor', cls):
            return tr('Expecting no return value when calling constructor of record `{cls}`.',
                    cls=cls)
    raise ValueError(f'Unexpected: {cn}')

def wrongReturnValue(ty: str) -> str:
    return tr('But the call returns a value of type `{ty}`.', ty=ty)

def unexpectedReturn(line: int) -> str:
    return tr('Call in line {line} causes the problematic return:', line=line)

def unexpectedNoReturn(line: int) -> str:
    return tr('Call in line {line} causes the function to return no value:', line=line)

def expectingReturnOfType(cn: location.CallableName, ty: str) -> str:
    match cn.kind:
        case 'function':
            return tr('Expecting return value of type `{ty}` when calling function `{fun}`.',
                    fun=cn.name, ty=ty)
        case location.ClassMember('method', cls):
            return tr('Expecting return value of type `{ty}` when calling method `{method}` of class `{cls}`.',
                    method=cn.name, cls=cls, ty=ty)
        case location.ClassMember('recordConstructor', cls):
            return tr('Expecting return value of type `{ty}` when calling constructor of record `{cls}`.',
                    cls=cls, ty=ty)
    raise ValueError(f'Unexpected: {cn}')

def noReturnValue() -> str:
    return tr('But no value returned.')

def transArg(pos: int):
    match getLang():
        case 'en':
            match pos:
                case 1: return '1st argument'
                case 2: return '2nd argument'
                case 3: return '3rd argument'
                case _: return f'{pos}th argument'
        case 'de':
            match pos:
                case 1: return 'erstes Argument'
                case 2: return 'zweites Argument'
                case 3: return 'drittes Argument'
                case 4: return 'viertes Argument'
                case 5: return 'fünftes Argument'
                case 6: return 'sechstes Argument'
                case 7: return 'siebtes Argument'
                case 8: return 'achtes Argument'
                case 9: return 'neuntes Argument'
                case _: return f'{pos}. Argument'
        case l:
            raise ValueError(f'Unexpected language: {l}')


def expectingArgumentOfTy(cn: location.CallableName, ty: str, pos: int | str) -> str:
    if isinstance(pos, int):
        arg = transArg(pos)
    else:
        match getLang():
            case 'en':
                arg = f'argument `{pos}`'
            case 'de':
                arg = f'Argument `{pos}`'
    match cn.kind:
        case 'function':
            return tr('The call of function `{fun}` expects value of type `{ty}` as {arg}.',
                    fun=cn.name, ty=ty, arg=arg)
        case location.ClassMember('method', cls):
            return tr('The call of method `{method}` of class `{cls}` expects value of type `{ty}` as {arg}.',
                    method=cn.name, cls=cls, ty=ty, arg=arg)
        case location.ClassMember('recordConstructor', cls):
            return tr('The call of the constructor of record `{cls}` expects value of type `{ty}` as {arg}.',
                    cls=cls, ty=ty, arg=arg)
    raise ValueError(f'Unexpected: {cn}')

def expectingDefaultValueOfTy(cn: location.CallableName, ty: str, paramName: str) -> str:
    match cn.kind:
        case 'function':
            return tr('Default value for parameter `{paramName}` of function `{fun}` must have type `{ty}`.',
                      paramName=paramName, fun=cn.name, ty=ty)
        case location.ClassMember('method', cls):
            return tr('Default value for parameter `{paramName}` of method `{method}` in class `{cls}` must have type `{ty}`.',
                      paramName=paramName, method=cn.name, cls=cls, ty=ty)
        case location.ClassMember('recordConstructor', cls):
            return tr('Default value for attribute `{paramName}` of record `{cls}` must have type `{ty}`.',
                      paramName=paramName, cls=cls, ty=ty)
    raise ValueError(f'Unexpected: {cn}')

def argCount(n: int) -> str:
    match getLang():
        case 'en':
            if n == 0:
                return 'no arguments'
            elif n == 1:
                return '1 argument'
            else:
                return f'{n} arguments'
        case 'de':
            if n == 0:
                return 'keine Argumente'
            elif n == 1:
                return '1 Argument'
            else:
                return f'{n} Argumente'

def argCountExact(cn: location.CallableName, expected: int) -> str:
    match cn.kind:
        case 'function':
            header = tr('Function `{fun}` takes ', fun=cn.name)
        case location.ClassMember('method', cls):
            header = tr('Method `{method}` of class `{cls}` takes ',
                      method=cn.name, cls=cls)
        case location.ClassMember('recordConstructor', cls):
            header = tr('Constructor of record `{cls}` takes ',
                        cls=cls)
        case _:
            raise ValueError(f'Unexpected: {cn}')
    return header + argCount(expected) + '.'

def argCountMin(cn: location.CallableName, expected: int) -> str:
    match cn.kind:
        case 'function':
            header = tr('Function `{fun}` takes at least ', fun=cn.name)
        case location.ClassMember('method', cls):
            header = tr('Method `{method}` of class `{cls}` takes at least ',
                      method=cn.name, cls=cls)
        case location.ClassMember('recordConstructor', cls):
            header = tr('Constructor of record `{cls}` takes at least ',
                        cls=cls)
        case _:
            raise ValueError(f'Unexpected: {cn}')
    return header + argCount(expected) + '.'

def argCountMax(cn: location.CallableName, expected: int) -> str:
    match cn.kind:
        case 'function':
            header = tr('Function `{fun}` takes at most ', fun=cn.name)
        case location.ClassMember('method', cls):
            header = tr('Method `{method}` of class `{cls}` takes at most ',
                      method=cn.name, cls=cls)
        case location.ClassMember('recordConstructor', cls):
            header = tr('Constructor of record `{cls}` takes at most ',
                        cls=cls)
        case _:
            raise ValueError(f'Unexpected: {cn}')
    return header + argCount(expected) + '.'

def realArgumentTy(ty: str) -> str:
    return tr('But the value given has type `{ty}`.', ty=ty)

def realDefaultValueTy(ty: str) -> str:
    return tr('But the default value has type `{ty}`.', ty=ty)

def realSetAttrTy(ty: str) -> str:
    return tr('Cannot set attribute to value of type `{ty}`.', ty=ty)

def expectingTypeAnnotation(cn: location.CallableName, param: str) -> str:
    match cn.kind:
        case 'function':
            return tr('Parameter `{param}` of function `{fun}` requires a type annotation.',
                        fun=cn.name, param=param)
        case location.ClassMember('method', cls):
            return tr('Parameter `{param}` of method `{method}` from class `{cls}` requires a type annotation.',
                        method=cn.name, cls=cls, param=param)
        case location.ClassMember('recordConstructor', cls):
            return tr('Parameter `{param}` of constructor of record `{cls}` requires a type annotation.',
                        cls=cls, param=param)
    raise ValueError(f'Unexpected: {cn}')

def noTypeAnnotationForAttribute(attrName: str, recordName: str) -> str:
    return tr('Attribute `{name}` of record `{record}` requires a type annotation.',
              name=attrName, record=recordName)

def invalidTy(ty: Any) -> str:
    return tr('invalid type `{ty}`', ty=ty)

def didYouMean(ty: str) -> str:
    return tr('Did you mean `{ty}`?', ty=ty)

def recordAttrDeclTy(recordName: str, attrName: str, ty: Any) -> str:
    return tr('Attribute `{attrName}` of record `{recordName}` declared with type `{ty}`.',
              recordName=recordName, attrName=attrName, ty=ty)

def unknownKeywordArgument(cn: location.CallableName, name: str) -> str:
    match cn.kind:
        case 'function':
            return tr('Function `{fun}` does not accept keyword argument `{name}`.',
                        fun=cn.name, name=name)
        case location.ClassMember('method', cls):
            return tr('Method `{method}` from class `{cls}` does not accept keyword argument `{name}`.',
                        method=cn.name, cls=cls, name=name)
        case location.ClassMember('recordConstructor', cls):
            return tr('Constructor of record `{cls}` does not accept keyword argument `{name}`.',
                        cls=cls, name=name)
    raise ValueError(f'Unexpected: {cn}')

def checkExpected(expected: str, actual: str) -> str:
    return tr('Expected {expected}, but the result is {actual}', expected=expected,actual=actual)

def numTests(n: int) -> str:
    match getLang():
        case 'en':
            if n == 0:
                return 'no tests'
            elif n == 1:
                return '1 test'
            else:
                return f'{n} tests'
        case 'de':
            if n == 0:
                return 'keine Tests'
            elif n == 1:
                return '1 Test'
            else:
                return f'{n} Tests'

def numFailing(n: int) -> str:
    match getLang():
        case 'en':
            if n == 0:
                return 'no errors'
            elif n == 1:
                return '1 error'
            else:
                return f'{n} errors'
        case 'de':
            if n == 0:
                return 'keine Fehler'
            elif n == 1:
                return '1 Fehler'
            else:
                return f'{n} Fehler'



import sys
import constants
sys.path.insert(0, constants.CODE_DIR)

import sys
import os

requiredVersion = (3, 12, 0)
def pythonVersionOk(v):
    (reqMajor, reqMinor, reqMicro) = requiredVersion
    if v.major < reqMajor or v.minor < reqMinor:
        return False
    if v.major == reqMajor and v.minor == reqMinor and v.micro < reqMicro:
        return False
    else:
        return True

if not pythonVersionOk(sys.version_info):
    vStr = sys.version.split()[0]
    reqVStr = '.'.join([str(x) for x in requiredVersion])
    print(f"""
Python in version {reqVStr} or newer is required. You are still using version {vStr}, please upgrade!
""")
    sys.exit(1)

# local imports
from constants import *
import i18n
import paths
from myLogging import *
import version as versionMod
import interactive
import runCode
import exceptionHandler
import cmdlineArgs


def printWelcomeString(file, version, doTypecheck):
    cwd = os.getcwd() + "/"
    if file.startswith(cwd):
        file = file[len(cwd):]
    versionStr = '' if not version else 'Version %s, ' % version
    pythonVersion = sys.version.split()[0]
    tycheck = ''
    if not doTypecheck:
        tycheck = ', no typechecking'
    printStderr(i18n.tr('=== WELCOME to ') + '"Write Your Python Program" ' +
                '(%sPython %s, %s%s) ===' % (versionStr, pythonVersion, file, tycheck))

def main(globals, argList=None):
    (args, restArgs) = cmdlineArgs.parseCmdlineArgs(argList)
    if args.verbose:
        enableVerbose()
    if args.debug:
        enableDebug()

    verbose(f'VERBOSE={args.verbose}, DEBUG={args.debug}')

    if args.lang:
        if args.lang in i18n.allLanguages:
            i18n.setLang(args.lang)
        else:
            printStderr(f'Unsupported language {args.lang}. Supported: ' + ', '.join(i18n.allLanguages))
            sys.exit(1)

    fileToRun: str = args.file
    if args.changeDir:
        os.chdir(os.path.dirname(fileToRun))
        fileToRun = os.path.basename(fileToRun)

    isInteractive = args.interactive
    version = versionMod.readVersion()
    if isInteractive:
        interactive.prepareInteractive(reset=not args.noClear)

    if fileToRun is None:
        return

    if not args.checkRunnable and (not args.quiet or args.verbose):
        printWelcomeString(fileToRun, version, doTypecheck=args.checkTypes)

    libDefs = runCode.prepareLib(onlyCheckRunnable=args.checkRunnable, enableTypeChecking=args.checkTypes)

    with (runCode.RunSetup(os.path.dirname(fileToRun), [fileToRun] + restArgs),
          paths.projectDir(os.path.abspath(os.getcwd()))):
        globals['__name__'] = '__wypp__'
        sys.modules['__wypp__'] = sys.modules['__main__']
        loadingFailed = False
        try:
            verbose(f'running code in {fileToRun}')
            globals['__file__'] = fileToRun
            globals = runCode.runStudentCode(fileToRun, globals, args.checkRunnable,
                                             doTypecheck=args.checkTypes, extraDirs=args.extraDirs)
        except Exception as e:
            verbose(f'Error while running code in {fileToRun}')
            exceptionHandler.handleCurrentException(exit=not isInteractive)
            loadingFailed = True

        runCode.performChecks(args.check, args.testFile, globals, libDefs, doTypecheck=args.checkTypes,
                              extraDirs=args.extraDirs, loadingFailed=loadingFailed)

        if args.repls:
            import replTester
            replTester.testRepls(args.repls, globals)

        if isInteractive:
            interactive.enterInteractive(globals, args.checkTypes, loadingFailed)





import unittest
import os
import location
from locationTestData import *

class TestLocation(unittest.TestCase):

    def assertLocation(self,
                       loc: location.Loc | None,
                       startLine: int, startCol: int, endLine: int, endCol: int):
        if loc is None:
            self.fail("loc is None")
        self.assertEqual('locationTestData.py', os.path.basename(loc.filename))
        self.assertEqual(startLine, loc.startLine)
        self.assertEqual(endLine, loc.endLine)
        self.assertEqual(startCol, loc.startCol)
        self.assertEqual(endCol, loc.endCol)

    def test_locationOfArgument(self):
        fi = myFun("foo", [1,2,3])
        loc = location.locationOfArgument(fi, 0)
        colNo = 21
        self.assertLocation(loc, lineNoMyFunCall,
                            colNo,
                            lineNoMyFunCall,
                            colNo + len("blub") + 2)

    def test_StdCallableInfo(self):
        info = location.StdCallableInfo(myFun, 'function')
        self.assertEqual('locationTestData.py', os.path.basename(info.file))
        resultLoc = info.getResultTypeLocation()
        argLoc = info.getParamSourceLocation("depth")
        self.assertLocation(resultLoc, lineNoMyFunDef, 49, lineNoMyFunDef, 66)
        self.assertLocation(argLoc, lineNoMyFunDef, 32, lineNoMyFunDef, 42)

    def test_StdCallableInfo2(self):
        info = location.StdCallableInfo(myFunNoResult, 'function')
        self.assertEqual('locationTestData.py', os.path.basename(info.file))
        resultLoc = info.getResultTypeLocation()
        self.assertIsNone(resultLoc)
        argLoc = info.getParamSourceLocation("depth")
        self.assertLocation(argLoc, lineNoMyFunNoResultDef, 18, lineNoMyFunNoResultDef, 28)

    def test_StdCallableInfoSub(self):
        sub = Sub()
        info = location.StdCallableInfo(sub.foo, location.ClassMember('method', 'Sub'))
        self.assertEqual('locationTestData.py', os.path.basename(info.file))
        argLoc = info.getParamSourceLocation("y")
        self.assertLocation(argLoc, lineFooSub, 18, lineFooSub, 24)

    def test_StdCallableInfoBase(self):
        b = Base()
        info = location.StdCallableInfo(b.foo, location.ClassMember('method', 'Base'))
        self.assertEqual('locationTestData.py', os.path.basename(info.file))
        argLoc = info.getParamSourceLocation("y")
        self.assertLocation(argLoc, lineFooBase, 26, lineFooBase, 32)

    def test_RecordConstructorInfo(self):
        info = location.RecordConstructorInfo(TestRecord)
        # Test getting parameter source locations with precise line/column numbers
        argLocX = info.getParamSourceLocation("x")
        argLocY = info.getParamSourceLocation("y")
        argLocZ = info.getParamSourceLocation("z")
        self.assertLocation(argLocX, lineNoTestRecordDef + 1, 4, lineNoTestRecordDef + 1, 10)
        self.assertLocation(argLocY, lineNoTestRecordDef + 2, 4, lineNoTestRecordDef + 2, 10)
        self.assertLocation(argLocZ, lineNoTestRecordDef + 3, 4, lineNoTestRecordDef + 3, 19)
        # Test non-existent parameter
        nonExistentLoc = info.getParamSourceLocation("nonexistent")
        self.assertIsNone(nonExistentLoc)
        # Test result type location (should be None for constructors)
        resultLoc = info.getResultTypeLocation()
        self.assertIsNone(resultLoc)



from __future__ import annotations
from collections.abc import Callable
from typing import ParamSpec, TypeVar, Any, Optional, Literal
import inspect
from dataclasses import dataclass
import utils
from myTypeguard import matchesTy, MatchesTyResult, MatchesTyFailure, Namespaces
import stacktrace
import location
import errors
from myLogging import *

def printVars(what: str, *l):
    s = what + ": " + ', '.join([str(x) for x in l])
    print(s)

def isEmptyAnnotation(t: Any) -> bool:
    return t is inspect.Signature.empty or t is inspect.Parameter.empty

def isEmptySignature(sig: inspect.Signature) -> bool:
    for x in sig.parameters:
         p = sig.parameters[x]
         if not isEmptyAnnotation(p.annotation):
             return False
    return isEmptyAnnotation(sig.return_annotation)

def handleMatchesTyResult(res: MatchesTyResult, tyLoc: Optional[location.Loc]) -> bool:
    match res:
        case MatchesTyFailure(exc, ty):
            if isDebug():
                debug(f'Exception occurred while calling matchesTy with type {ty}, re-raising')
                raise exc
            else:
                raise errors.WyppTypeError.invalidType(ty, tyLoc)
        case b:
            return b

def getKind(cfg: CheckCfg, paramNames: list[str]) -> Literal['function', 'method', 'staticmethod']:
    kind: Literal['function', 'method', 'staticmethod'] = 'function'
    match cfg.kind:
        case location.ClassMember('recordConstructor', _):
            kind = 'method'
        case location.ClassMember('method', _):
            if len(paramNames) > 0 and paramNames[0] == 'self':
                kind = 'method'
            else:
                kind = 'staticmethod'
        case 'function':
            kind = 'function'
    return kind

def checkSignature(sig: inspect.Signature, info: location.CallableInfo, cfg: CheckCfg) -> None:
    paramNames = list(sig.parameters)
    kind = getKind(cfg, paramNames)
    for i in range(len(paramNames)):
        name = paramNames[i]
        p = sig.parameters[name]
        ty = p.annotation
        if isEmptyAnnotation(ty):
            if i == 0 and kind == 'method':
                pass
            else:
                locDecl = info.getParamSourceLocation(name)
                raise errors.WyppTypeError.partialAnnotationError(location.CallableName.mk(info), name, locDecl)
        if p.default is not inspect.Parameter.empty:
            locDecl = info.getParamSourceLocation(name)
            if not handleMatchesTyResult(matchesTy(p.default, ty, cfg.ns), locDecl):
                raise errors.WyppTypeError.defaultError(location.CallableName.mk(info), name, locDecl, ty, p.default)

def mandatoryArgCount(sig: inspect.Signature) -> int:
    required_kinds = {
        inspect.Parameter.POSITIONAL_ONLY,
        inspect.Parameter.POSITIONAL_OR_KEYWORD,
        inspect.Parameter.KEYWORD_ONLY,
    }
    res = 0
    for p in sig.parameters.values():
        if p.kind in required_kinds and p.default is inspect._empty:
            res = res + 1
    return res

def checkArgument(p: inspect.Parameter, name: str, idx: Optional[int], a: Any,
                  locArg: Optional[location.Loc], info: location.CallableInfo, cfg: CheckCfg):
    t = p.annotation
    if not isEmptyAnnotation(t):
        locDecl = info.getParamSourceLocation(name)
        if not handleMatchesTyResult(matchesTy(a, t, cfg.ns), locDecl):
            cn = location.CallableName.mk(info)
            raise errors.WyppTypeError.argumentError(cn,
                                                     name,
                                                     idx,
                                                     locDecl,
                                                     t,
                                                     a,
                                                     locArg)
def checkArguments(sig: inspect.Signature, args: tuple, kwargs: dict,
                   info: location.CallableInfo, cfg: CheckCfg) -> None:
    debug(f'Checking arguments when calling {info}')
    paramNames = list(sig.parameters)
    mandatory = mandatoryArgCount(sig)
    kind = getKind(cfg, paramNames)
    offset = 1 if kind == 'method' else 0
    fi = stacktrace.callerOutsideWypp()
    callLoc = None if not fi else location.Loc.fromFrameInfo(fi)
    cn = location.CallableName.mk(info)
    def raiseArgMismatch():
        raise errors.WyppTypeError.argCountMismatch(cn,
                                                    callLoc,
                                                    len(paramNames) - offset,
                                                    mandatory - offset,
                                                    len(args) - offset)
    if len(args) + len(kwargs) < mandatory:
        raiseArgMismatch()
    for i in range(len(args)):
        if i >= len(paramNames):
            raiseArgMismatch()
        name = paramNames[i]
        p = sig.parameters[name]
        locArg = None if fi is None else location.locationOfArgument(fi, i)
        checkArgument(p, name, i - offset, args[i], locArg, info, cfg)
    for name in kwargs:
        if name not in sig.parameters:
            raise errors.WyppTypeError.unknownKeywordArgument(cn, callLoc, name)
        locArg = None if fi is None else location.locationOfArgument(fi, name)
        checkArgument(sig.parameters[name], name, None, kwargs[name], locArg, info, cfg)

def checkReturn(sig: inspect.Signature, returnFrame: Optional[inspect.FrameInfo],
                result: Any, info: location.CallableInfo, cfg: CheckCfg) -> None:
    t = sig.return_annotation
    if isEmptyAnnotation(t):
        t = None
    debug(f'Checking return value when calling {info}, return type: {t}')
    locDecl = info.getResultTypeLocation()
    if not handleMatchesTyResult(matchesTy(result, t, cfg.ns), locDecl):
        fi = stacktrace.callerOutsideWypp()
        if fi is not None:
            locRes = location.Loc.fromFrameInfo(fi)
        returnLoc = None
        extraFrames = []
        if returnFrame:
            returnLoc = location.Loc.fromFrameInfo(returnFrame)
            extraFrames = [returnFrame]
        raise errors.WyppTypeError.resultError(location.CallableName.mk(info), locDecl, t, returnLoc, result,
                                               locRes, extraFrames)


@dataclass
class CheckCfg:
    kind: location.CallableKind
    ns: Namespaces
    @staticmethod
    def fromDict(d: dict) -> CheckCfg:
        k = d['kind']
        match k:
            case 'function':
                kind = 'function'
            case 'method':
                kind = location.ClassMember('method', d['className'])
        return CheckCfg(kind=kind, ns=Namespaces(d['globals'], d['locals']))

P = ParamSpec("P")
T = TypeVar("T")

def wrapTypecheck(cfg: dict | CheckCfg, outerInfo: Optional[location.CallableInfo]=None) -> Callable[[Callable[P, T]], Callable[P, T]]:
    if isinstance(cfg, CheckCfg):
        checkCfg = cfg
    else:
        checkCfg = CheckCfg.fromDict(cfg)
    def _wrap(f: Callable[P, T]) -> Callable[P, T]:
        sig = inspect.signature(f)
        if isEmptySignature(sig):
            return f
        if outerInfo is None:
            # we have a regular function or method
            info = location.StdCallableInfo(f, checkCfg.kind)
        else:
            # special case: constructor of a record
            info = outerInfo
        utils._call_with_frames_removed(checkSignature, sig, info, checkCfg)
        def wrapped(*args, **kwargs) -> T:
            utils._call_with_frames_removed(checkArguments, sig, args, kwargs, info, checkCfg)
            returnTracker = stacktrace.getReturnTracker()
            result = utils._call_with_next_frame_removed(f, *args, **kwargs)
            retFrame = returnTracker.getReturnFrame(0)
            utils._call_with_frames_removed(
                checkReturn, sig, retFrame, result, info, checkCfg
            )
            return result
        return wrapped
    return _wrap

def wrapTypecheckRecordConstructor(cls: type, ns: Namespaces) -> Callable:
    checkCfg = CheckCfg.fromDict({'kind': 'method', 'className': cls.__name__,
                                  'globals': ns.globals, 'locals': ns.locals})
    info = location.RecordConstructorInfo(cls)
    return wrapTypecheck(checkCfg, info)(cls.__init__)



import unittest
import i18n
import location
import inspect
import utils

class TestI18nTranslations(unittest.TestCase):
    def setUp(self):
        # Get all functions that return strings
        blacklist = ['tr']
        self.stringFunctions = []
        for name, func in inspect.getmembers(i18n, inspect.isfunction):
            if not name in blacklist and \
                hasattr(func, '__annotations__') and func.__annotations__.get('return') == str:
                self.stringFunctions.append((name, func))

    def _getCallableNameVariants(self):
        return [
            location.CallableName('foo', 'function'),
            location.CallableName('foo', location.ClassMember('method', 'foo')),
            location.CallableName('foo', location.ClassMember('recordConstructor', 'foo'))
        ]

    def _getTestArgs(self, func):
        sig = inspect.signature(func)
        args = []
        hasCallableName = False

        for paramName, param in sig.parameters.items():
            if param.annotation == str:
                args.append('foo')
            elif param.annotation == int:
                args.append(1)
            elif param.annotation == location.CallableName:
                args.append(None)  # Placeholder
                hasCallableName = True
            else:
                args.append('foo')  # Default fallback

        if hasCallableName:
            # Generate variants for each CallableName parameter
            variants = []
            for variant in self._getCallableNameVariants():
                testArgs = []
                for i, arg in enumerate(args):
                    if arg is None:  # CallableName placeholder
                        testArgs.append(variant)
                    else:
                        testArgs.append(arg)
                variants.append(testArgs)
            return variants
        else:
            return [args] if args else [[]]

    def testAllI18NFunctions(self):
        with utils.underTest(True):
            for lang in i18n.allLanguages:
                with i18n.explicitLang(lang):
                    with self.subTest(lang=lang):
                        for funcName, func in self.stringFunctions:
                            with self.subTest(function=funcName):
                                argVariants = self._getTestArgs(func)
                                for args in argVariants:
                                    with self.subTest(args=args):
                                        result = func(*args)
                                        self.assertIsInstance(result, str)
                                        self.assertGreater(len(result), 0)

if __name__ == '__main__':
    unittest.main()



import sys
import doctest
from myLogging import *

# We use our own DocTestParser to replace exception names in stacktraces


def rewriteLines(lines: list[str]):
    """
    Each line has exactly one of the following four kinds:
    - COMMENT: if it starts with '#' (leading whitespace stripped)
    - PROMPT: if it starts with '>>>' (leading whitespace stripped)
    - EMPTY: if it contains only whitespace
    - OUTPUT: otherwise

    rewriteLines replaces every EMPTY lines with '<BLANKLINE>', provided
    the first non-EMPTY line before the line has kind PROMPT OR OUTPUT
    and the next non-EMPTY line after the line has kind OUTPUT.
    """

    def get_line_kind(line: str) -> str:
        stripped = line.lstrip()
        if not stripped:
            return 'EMPTY'
        elif stripped.startswith('#'):
            return 'COMMENT'
        elif stripped.startswith('>>>'):
            return 'PROMPT'
        else:
            return 'OUTPUT'

    def find_prev_non_empty(idx: int) -> tuple[int, str]:
        """Find the first non-EMPTY line before idx. Returns (index, kind)"""
        for i in range(idx - 1, -1, -1):
            kind = get_line_kind(lines[i])
            if kind != 'EMPTY':
                return i, kind
        return -1, 'NONE'

    def find_next_non_empty(idx: int) -> tuple[int, str]:
        """Find the first non-EMPTY line after idx. Returns (index, kind)"""
        for i in range(idx + 1, len(lines)):
            kind = get_line_kind(lines[i])
            if kind != 'EMPTY':
                return i, kind
        return -1, 'NONE'

    # Process each line
    for i in range(len(lines)):
        if get_line_kind(lines[i]) == 'EMPTY':
            # Check conditions for replacement
            prev_idx, prev_kind = find_prev_non_empty(i)
            next_idx, next_kind = find_next_non_empty(i)

            # Replace if previous is PROMPT or OUTPUT and next is OUTPUT
            if prev_kind in ['PROMPT', 'OUTPUT'] and next_kind == 'OUTPUT':
                lines[i] = '<BLANKLINE>'


class MyDocTestParser(doctest.DocTestParser):
    def get_examples(self, string, name='<string>'):
        """
        The string is the docstring from the file which we want to test.
        """
        prefs = {'WyppTypeError: ': 'errors.WyppTypeError: ',
                 'WyppNameError: ': 'errors.WyppNameError: ',
                 'WyppAttributeError: ': 'errors.WyppAttributeError: '}
        lines = []
        for l in string.split('\n'):
            for pref,repl in prefs.items():
                if l.startswith(pref):
                    l = repl + l
            lines.append(l)
        rewriteLines(lines)
        string = '\n'.join(lines)
        x = super().get_examples(string, name)
        return x

def testRepl(repl: str, defs: dict) -> tuple[int, int]:
    doctestOptions = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    (failures, tests) = doctest.testfile(repl, globs=defs, module_relative=False,
                                         optionflags=doctestOptions, parser=MyDocTestParser())
    if failures == 0:
        if tests == 0:
            print(f'No tests in {repl}')
        else:
            print(f'All {tests} tests in {repl} succeeded')
    else:
        print(f'ERROR: {failures} out of {tests} in {repl} failed')
    return (failures, tests)

def testRepls(repls: list[str], defs: dict):
    totalFailures = 0
    totalTests = 0
    for r in repls:
        (failures, tests) = testRepl(r, defs)
        totalFailures += failures
        totalTests += tests

    if totalFailures == 0:
        if totalTests == 0:
            print('ERROR: No tests found at all!')
            sys.exit(1)
        else:
            print(f'All {totalTests} tests succeeded. Great!')
    else:
        print(f'ERROR: {failures} out of {tests} failed')
        sys.exit(1)
